# fmt: off
import torch.nn as nn
import torch.optim as optim
from timm.optim import Lamb


# fmt:on
def optim_cifar10_resnet18(model: nn.Module) -> dict:
    """optimizer to train a ResNet18 on CIFAR-10"""
    optimizer = optim.SGD(
        model.parameters(),
        lr=0.05,
        momentum=0.9,
        weight_decay=5e-4,
    )
    scheduler = optim.lr_scheduler.MultiStepLR(
        optimizer,
        milestones=[25, 50],
        gamma=0.1,
    )
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def optim_cifar10_resnet50(model: nn.Module) -> dict:
    r"""Hyperparameters from Deep Residual Learning for Image Recognition
    https://arxiv.org/pdf/1512.03385.pdf
    """
    optimizer = optim.SGD(
        model.parameters(),
        lr=0.1,
        momentum=0.9,
        weight_decay=5e-4,
        nesterov=True,
    )
    scheduler = optim.lr_scheduler.MultiStepLR(
        optimizer,
        milestones=[60, 120, 160],
        gamma=0.2,
    )
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def optim_cifar10_wideresnet(model: nn.Module) -> dict:
    """optimizer to train a WideResNet28x10 on CIFAR-10"""
    optimizer = optim.SGD(
        model.parameters(),
        lr=0.1,
        momentum=0.9,
        weight_decay=5e-4,
        nesterov=True,
    )
    scheduler = optim.lr_scheduler.MultiStepLR(
        optimizer,
        milestones=[60, 120, 160],
        gamma=0.2,
    )
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def optim_cifar100_resnet18(model: nn.Module) -> dict:
    optimizer = optim.SGD(
        model.parameters(),
        lr=0.1,
        momentum=0.9,
        weight_decay=1e-4,
        nesterov=True,
    )
    scheduler = optim.lr_scheduler.MultiStepLR(
        optimizer,
        milestones=[60, 120, 160],
        gamma=0.2,
    )
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def optim_cifar100_resnet50(model: nn.Module, adam: bool = False) -> dict:
    r"""Hyperparameters from Deep Residual Learning for Image Recognition
    https://arxiv.org/pdf/1512.03385.pdf
    """
    optimizer = optim.SGD(
        model.parameters(),
        lr=0.1,
        momentum=0.9,
        weight_decay=5e-4,
        nesterov=True,
    )
    scheduler = optim.lr_scheduler.MultiStepLR(
        optimizer,
        milestones=[60, 120, 160],
        gamma=0.2,
    )
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def optim_imagenet_resnet50(
    model: nn.Module,
    n_epochs: int = 90,
    start_lr: float = 0.256,
    end_lr: float = 0,
) -> dict:
    r"""Hyperparameters from Deep Residual Learning for Image Recognition
    https://arxiv.org/pdf/1512.03385.pdf
    """
    optimizer = optim.SGD(
        model.parameters(),
        lr=start_lr,
        momentum=0.875,
        weight_decay=3.0517578125e-05,
        nesterov=False,
    )
    scheduler = optim.lr_scheduler.CosineAnnealingLR(
        optimizer, n_epochs, eta_min=end_lr
    )
    return {
        "optimizer": optimizer,
        "lr_scheduler": scheduler,
        "monitor": "hp/val_acc",
    }


def optim_imagenet_resnet50_A3(
    model: nn.Module, effective_batch_size: int = None
) -> dict:
    """
    Training procedure proposed in ResNet strikes back: An improved training
        procedure in timm.

    Args:
        model (nn.Module): The model to be optimized.
        effective_batch_size (int, optional): The batch size of the model
            (taking multiple GPUs into account). Defaults to None.

    Returns:
        dict: The optimizer and the scheduler for the training.
    """
    if effective_batch_size is None:
        print("Setting effective batch size to 2048 for steps computations !")
        effective_batch_size = 2048

    optimizer = Lamb(model.parameters(), lr=0.008, weight_decay=0.02)

    warmup = optim.lr_scheduler.LinearLR(
        optimizer,
        start_factor=1e-4,
        end_factor=1,
        total_iters=5 * (1281167 // effective_batch_size + 1),
    )
    cosine_scheduler = optim.lr_scheduler.CosineAnnealingLR(
        optimizer,
        eta_min=1e-6,
        T_max=105 * (1281167 // effective_batch_size + 1),
    )
    scheduler = optim.lr_scheduler.SequentialLR(
        optimizer,
        schedulers=[warmup, cosine_scheduler],
        milestones=[5 * (1281167 // effective_batch_size + 1)],
    )
    return {
        "optimizer": optimizer,
        "lr_scheduler": {
            "scheduler": scheduler,
            "interval": "step",
            "frequency": 1,
        },
        "monitor": "hp/val_acc",
    }



# fmt: off
from argparse import ArgumentParser
from pathlib import Path
from typing import Any, List, Optional, Union

import torchvision.transforms as T
from pytorch_lightning import LightningDataModule
from timm.data.auto_augment import rand_augment_transform
from torch.utils.data import DataLoader, Dataset, random_split
from torchvision.datasets import DTD, SVHN, ImageNet, INaturalist

from ..datasets import ImageNetO, ImageNetR


# fmt: on
class ImageNetDataModule(LightningDataModule):
    def __init__(
        self,
        root: Union[str, Path],
        batch_size: int,
        val_split: int = 0,
        num_workers: int = 1,
        pin_memory: bool = True,
        persistent_workers: bool = True,
        ood_name: str = "svhn",
        test_option: str = None,
        # num_ops: int = 2,
        # magnitude: float = 15,
        procedure: str = "A3",
        **kwargs,
    ) -> None:
        super().__init__()

        if isinstance(root, str):
            root = Path(root)

        self.root: Path = root
        self.batch_size = batch_size
        self.val_split = val_split
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.persistent_workers = persistent_workers
        self.ood_name = ood_name
        self.num_classes = 1000
        self.num_channels = 3

        if test_option is None:
            self.dataset = ImageNet
        elif test_option == "imagenet-r":
            self.dataset = ImageNetR
        else:
            raise ValueError(f"Error, {test_option} not taken in charge.")

        if ood_name == "inaturalist":
            self.ood_dataset = INaturalist
        elif ood_name == "imagenet-o":
            self.ood_dataset = ImageNetO
        elif ood_name == "svhn":
            self.ood_dataset = SVHN
        elif ood_name == "textures":
            self.ood_dataset = DTD
        else:
            raise ValueError(f"The dataset {ood_name} is not supported.")

        self.procedure = procedure

        if self.procedure == "Classic":
            print("Classic Procedure")
            train_size = 224
        elif self.procedure == "A3":
            print("Procedure A3")
            train_size = 160
        else:
            raise ValueError("The procedure is unknown")

        self.transform_train = T.Compose(
            [
                T.RandomResizedCrop(train_size),
                T.RandomHorizontalFlip(),
                rand_augment_transform("rand-m6-mstd0.5-inc1", {}),
                T.ToTensor(),
                T.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
            ]
        )

        self.transform_test = T.Compose(
            [
                T.Resize(256),
                T.CenterCrop(224),
                T.ToTensor(),
                T.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
            ]
        )

    def _verify_splits(self, split: str) -> None:
        if split not in list(self.root.iterdir()):
            raise FileNotFoundError(
                f"a {split} Imagenet split was not found in {self.root},"
                f" make sure the folder contains a subfolder named {split}"
            )

    def prepare_data(self) -> None:
        if self.ood_name == "inaturalist":
            self.ood = self.ood_dataset(
                self.root,
                version="2021_valid",
                download=True,
                transform=self.transform_test,
            )
        elif self.ood_name != "textures":
            self.ood = self.ood_dataset(
                self.root,
                split="test",
                download=True,
                transform=self.transform_test,
            )
        else:
            self.ood = self.ood_dataset(
                self.root,
                split="train",
                download=True,
                transform=self.transform_test,
            )

    def setup(self, stage: Optional[str] = None) -> None:
        if stage == "fit" or stage is None:
            full = self.dataset(
                self.root,
                split="train",
                transform=self.transform_train,
            )
            self.train, self.val = random_split(
                full, [len(full) - self.val_split, self.val_split]
            )
            if self.val_split == 0:
                self.val = self.dataset(
                    self.root,
                    split="val",
                    transform=self.transform_test,
                )
        if stage == "test" or stage is None:
            self.test = self.dataset(
                self.root,
                split="val",
                transform=self.transform_test,
            )
            if self.ood_name == "inaturalist":
                self.ood = self.ood_dataset(
                    self.root,
                    version="2021_valid",
                    transform=self.transform_test,
                )
            else:
                self.ood = self.ood_dataset(
                    self.root,
                    transform=self.transform_test,
                    download=True,
                )

    def train_dataloader(self) -> DataLoader:
        return self._data_loader(self.train, shuffle=True)

    def val_dataloader(self) -> DataLoader:
        return self._data_loader(self.val)

    def test_dataloader(self) -> List[DataLoader]:
        r"""Gets test dataloaders for ImageNet.
        Returns:
            List[DataLoader]: ImageNet test set (in distribution data) and SVHN
                test split (out-of-distribution data).
        """
        return [self._data_loader(self.test), self._data_loader(self.ood)]

    def _data_loader(
        self, dataset: Dataset, shuffle: bool = False
    ) -> DataLoader:
        return DataLoader(
            dataset,
            batch_size=self.batch_size,
            shuffle=shuffle,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            persistent_workers=self.persistent_workers,
        )

    @classmethod
    def add_argparse_args(
        cls,
        parent_parser: ArgumentParser,
        **kwargs: Any,
    ) -> ArgumentParser:
        p = parent_parser.add_argument_group("datamodule")
        p.add_argument("--root", type=str, default="./data/")
        p.add_argument("--batch_size", type=int, default=256)
        p.add_argument("--val_split", type=int, default=0)
        p.add_argument("--num_workers", type=int, default=4)
        p.add_argument("--ood_name", type=str, default="svhn")
        p.add_argument("--test_option", type=str)
        return parent_parser



# fmt: off
from pathlib import Path
from typing import Tuple, Union


# fmt: on
def get_version(
    root: Union[str, Path], version: int, checkpoint: int = None
) -> Tuple[Path, Path]:
    """
    Find a compute the path to the checkpoint corresponding to the input
        parameters


    Args:
        root (Union[str, Path]): The root of the dataset containing the
            checkpoints.
        version (int): The version of the checkpoint
        checkpoint (int, optional): The number of the checkpoint. Defaults
            to None.

    Raises:
        Exception: if the checkpoint cannot be found.

    Returns:
        Tuple[Path, Path]: The path to the checkpoints and to its parameters.
    """
    if isinstance(root, str):
        root = Path(root)

    if (root / f"version_{version}").is_dir():
        version_folder = root / f"version_{version}"
        ckpt_folder = version_folder / "checkpoints"
        if checkpoint is None:
            ckpts = list(ckpt_folder.glob("*.ckpt"))
        else:
            ckpts = list(ckpt_folder.glob(f"epoch={checkpoint}-*.ckpt"))
    else:
        raise Exception(
            f"The directory {root}/version_{version} does not exist."
        )

    file = ckpts[0]
    return (file.resolve(), (version_folder / "hparams.yaml").resolve())



# fmt:off
from typing import Any, Literal, Optional

import torch
import torch.nn.functional as F
from torch import Tensor
from torchmetrics import Metric
from torchmetrics.utilities import rank_zero_warn
from torchmetrics.utilities.data import dim_zero_cat


# fmt:on
class JensenShannonDivergence(Metric):
    """The Jensen Shannon Divergence Metric to estimate the epistemic
    uncertainty of an ensemble of estimators.

    Args:
        reduction (str, optional): Determines how to reduce over the
            :math:`B`/batch dimension:

            - ``'mean'`` [default]: Averages score across samples
            - ``'sum'``: Sum score across samples
            - ``'none'`` or ``None``: Returns score per sample

        kwargs: Additional keyword arguments, see `Advanced metric settings
            <https://torchmetrics.readthedocs.io/en/stable/pages/overview.html#metric-kwargs>`_.

    Inputs:
        - :attr:`probs`: :math:`(B, N, C)`

        where :math:`B` is the batch size, :math:`C` is the number of classes
        and :math:`N` is the number of estimators.

    Note:
        A higher Jensen-Shannon divergence means a higher uncertainty.

    Warning:
        Make sure that the probabilities in :attr:`probs` are normalized to sum
        to one.

    Raises:
        ValueError:
            If :attr:`reduction` is not one of ``'mean'``, ``'sum'``,
            ``'none'`` or ``None``.
    """

    full_state_update: bool = False
    higher_is_better: Optional[bool] = None
    full_state_update: bool = False

    def __init__(
        self,
        reduction: Literal["mean", "sum", "none", None] = "mean",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        allowed_reduction = ("sum", "mean", "none", None)
        if reduction not in allowed_reduction:
            raise ValueError(
                "Expected argument `reduction` to be one of ",
                f"{allowed_reduction} but got {reduction}",
            )

        self.reduction = reduction

        if self.reduction in ["mean", "sum"]:
            self.add_state(
                "values",
                default=torch.tensor(0.0),
                dist_reduce_fx="sum",
            )
        else:
            self.add_state("values", default=[], dist_reduce_fx="cat")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, probs: Tensor) -> None:  # type: ignore
        """Update state with new data.

        Args:
            probs (torch.Tensor): Probabilities from the model.
        """
        batch_size = probs.size(0)
        num_estimators = probs.size(1)
        mean_probs = probs.mean(dim=1, keepdim=True).repeat(
            1, probs.shape[1], 1
        )
        jsd = (
            F.kl_div(
                mean_probs.log(),
                probs.log(),
                log_target=True,
                reduction="none",
            ).sum(dim=(1, 2))
            / num_estimators
        )

        if self.reduction is None or self.reduction == "none":
            self.values.append(jsd)
        else:
            self.values += jsd.sum()
            self.total += batch_size

    def compute(self) -> Tensor:
        """Compute Jensen Shannon Divergence based on inputs passed in to
        ``update``.
        """
        values = dim_zero_cat(self.values)
        if self.reduction == "sum":
            return values.sum(dim=-1)
        elif self.reduction == "mean":
            return values.sum(dim=-1) / self.total
        else:  # reduction is None or "none"
            return values


class JensenShannonDivergence_old(Metric):
    full_state_update: bool = False

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.add_state("probs", [], dist_reduce_fx="cat")

        rank_zero_warn(
            "Metric `JensenShannonDivergence` will save all "
            "predictions in buffer. For large datasets this may lead to large "
            "memory footprint."
        )

    def update(self, probs: Tensor) -> None:  # type: ignore
        # store data as (example, estimator, class)
        self.probs.append(probs.transpose(0, 1))

    def compute(self) -> Tensor:
        probs = dim_zero_cat(self.probs)
        mean_proba = probs.mean(1, keepdim=True).repeat(1, probs.shape[1], 1)

        return (
            F.kl_div(
                mean_proba.log(),
                probs.log(),
                log_target=True,
                reduction="batchmean",
            )
            / probs.shape[1]
        )


if __name__ == "__main__":
    import pytest

    input = torch.softmax(torch.randn(8, 5, 4), dim=-1)  # (B, N, C)

    mi_old = JensenShannonDivergence_old()
    mi = JensenShannonDivergence()

    print(mi_old(input.transpose(0, 1)))
    print(mi(input))

    assert mi_old(input.transpose(0, 1)) == pytest.approx(mi(input))

    print("All good!")



from pathlib import Path
from typing import Any, Callable, Optional, Tuple

from torchvision.datasets import ImageFolder


class ImageNetO(ImageFolder):
    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        download: bool = False,
        split: str = None,
    ):
        super().__init__(root=root / Path("imagenet-o"), transform=transform)

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        return super().__getitem__(index)

    def __len__(self) -> int:
        return len(self.samples)



# fmt: off
from typing import Any, Literal, Optional

import torch
import torch.nn.functional as F
from torchmetrics import Metric
from torchmetrics.utilities import rank_zero_warn
from torchmetrics.utilities.data import dim_zero_cat


# fmt: on
class NegativeLogLikelihood_old(Metric):
    full_state_update: bool = False

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.add_state("log_probs", [], dist_reduce_fx="cat")
        self.add_state("targets", [], dist_reduce_fx="cat")

        rank_zero_warn(
            "Metric `NLLMetric` will save all targets and predictions"
            " in buffer. For large datasets this may lead to large memory"
            " footprint."
        )

    def update(self, probs: torch.Tensor, target: torch.Tensor) -> None:
        self.log_probs.append(torch.log(probs))
        self.targets.append(target)

    def compute(self) -> torch.Tensor:
        log_probs = dim_zero_cat(self.log_probs)
        targets = dim_zero_cat(self.targets)
        return F.nll_loss(log_probs, targets)


class NegativeLogLikelihood(Metric):
    """The Negative Log Likelihood Metric.

    Args:
        reduction (str, optional): Determines how to reduce over the
            :math:`B`/batch dimension:

            - ``'mean'`` [default]: Averages score across samples
            - ``'sum'``: Sum score across samples
            - ``'none'`` or ``None``: Returns score per sample

        kwargs: Additional keyword arguments, see `Advanced metric settings
            <https://torchmetrics.readthedocs.io/en/stable/pages/overview.html#metric-kwargs>`_.

    Inputs:
        - :attr:`probs`: :math:`(B, C)`
        - :attr:`target`: :math:`(B)`

        where :math:`B` is the batch size and :math:`C` is the number of
        classes.

    Warning:
        Make sure that the probabilities in :attr:`probs` are normalized to sum
        to one.

    Raises:
        ValueError:
            If :attr:`reduction` is not one of ``'mean'``, ``'sum'``,
            ``'none'`` or ``None``.
    """

    is_differentiable: bool = False
    higher_is_better: Optional[bool] = False
    full_state_update: bool = False

    def __init__(
        self,
        reduction: Literal["mean", "sum", "none", None] = "mean",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        allowed_reduction = ("sum", "mean", "none", None)
        if reduction not in allowed_reduction:
            raise ValueError(
                "Expected argument `reduction` to be one of ",
                f"{allowed_reduction} but got {reduction}",
            )

        self.reduction = reduction

        if self.reduction in ["mean", "sum"]:
            self.add_state(
                "values",
                default=torch.tensor(0.0),
                dist_reduce_fx="sum",
            )
        else:
            self.add_state("values", default=[], dist_reduce_fx="cat")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, probs: torch.Tensor, target: torch.Tensor) -> None:
        """Update state with prediction probabilities and targets.

        Args:
            probs (torch.Tensor): Probabilities from the model.
            target (torch.Tensor): Ground truth labels.
        """
        if self.reduction is None or self.reduction == "none":
            self.values.append(
                F.nll_loss(torch.log(probs), target, reduction="none")
            )
        else:
            self.values += F.nll_loss(torch.log(probs), target, reduction="sum")
            self.total += target.size(0)

    def compute(self) -> torch.Tensor:
        """Computes NLL based on inputs passed in to ``update`` previously."""
        values = dim_zero_cat(self.values)

        if self.reduction == "sum":
            return values.sum(dim=-1)
        elif self.reduction == "mean":
            return values.sum(dim=-1) / self.total
        else:  # reduction is None or "none"
            return values


if __name__ == "__main__":
    input = torch.softmax(torch.randn(8, 4), dim=-1)
    target = torch.randint(4, (8,))

    nll = NegativeLogLikelihood_old()
    nllv2 = NegativeLogLikelihood()

    assert nll(input, target) == nllv2(input, target)

    print("All good!")



# fmt:off
from pathlib import Path
from typing import Any, Callable, Optional, Tuple

from torchvision.datasets import VisionDataset

import numpy as np

# fmt:on


class CIFAR10_C(VisionDataset):
    cifarc_subsets = [
        "brightness",
        "contrast",
        "defocus_blur",
        "elastic_transform",
        "fog",
        "frost",
        "gaussian_blur",
        "gaussian_noise",
        "glass_blur",
        "impulse_noise",
        "jpeg_compression",
        "motion_blur",
        "pixelate",
        "saturate",
        "shot_noise",
        "snow",
        "spatter",
        "speckle_noise",
        "zoom_blur",
    ]

    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        dataset_path: Path = Path("CIFAR-10-C"),
        subset: str = "brightness",
        severity: int = 1,
    ):
        super().__init__(root=root / dataset_path, transform=transform)
        assert (
            subset in ["all"] + self.cifarc_subsets
        ), f"The subset '{subset}' does not exist in CIFAR-C."
        self.subset = subset
        self.severity = severity

        self.transform = transform
        self.target_transform = target_transform

        assert severity in list(
            range(1, 6)
        ), "Corruptions severity should be chosen between 1 and 5 included."
        samples, labels = self.make_dataset(
            self.root, self.subset, self.severity
        )

        self.samples = samples
        self.labels = labels

    def make_dataset(
        self, root: Path, subset: str, severity: int
    ) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Build the corrupted dataset according to the chosen subset and
            severity. If the subset is 'all', gather all corruption types
            in the dataset.
        Args:
            root (Path):The path to the dataset.
            subset (str): The name of the corruption subset to be used. Choose
                `all` for the dataset to contain all subsets.
            severity (int): The severity of the corruption applied to the
                images.
        Returns:
            Tuple[np.ndarray, np.ndarray]: The samples and labels of the chosen
        """
        if subset == "all":
            sample_arrays = []
            labels: np.ndarray = np.load(root / "labels.npy")[
                (severity - 1) * 10000 : severity * 10000
            ]
            for cifar_subset in self.cifarc_subsets:
                sample_arrays.append(
                    np.load(root / (cifar_subset + ".npy"))[
                        (severity - 1) * 10000 : severity * 10000
                    ]
                )
            samples = np.concatenate(sample_arrays, axis=0)
            labels = np.tile(labels, len(self.cifarc_subsets))

        else:
            samples: np.ndarray = np.load(root / (subset + ".npy"))[
                (severity - 1) * 10000 : severity * 10000
            ]
            labels: np.ndarray = np.load(root / "labels.npy")[
                (severity - 1) * 10000 : severity * 10000
            ]
        return samples, labels

    def __len__(self) -> int:
        return self.labels.shape[0]

    def __getitem__(self, index: int) -> Any:
        sample, target = (
            self.samples[index],
            self.labels[index],
        )

        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return sample, target


class CIFAR100_C(CIFAR10_C):
    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        dataset_path: Path = Path("CIFAR-100-C"),
        subset: str = "brightness",
        severity: int = 1,
    ):
        super().__init__(
            root,
            transform=transform,
            dataset_path=dataset_path,
            subset=subset,
            severity=severity,
        )



# fmt:off
import os
from collections import defaultdict
from typing import Callable, Optional

from PIL import Image
from torch.utils.data import Dataset

import numpy as np


# fmt:on
class TinyImageNetPaths:
    """From https://gist.github.com/z-a-f/b862013c0dc2b540cf96a123a6766e54."""

    def __init__(self, root):
        train_path = root / "train"
        val_path = root / "val"
        test_path = root / "test"

        wnids_path = root / "wnids.txt"
        words_path = root / "words.txt"

        self._make_paths(
            train_path, val_path, test_path, wnids_path, words_path
        )

    def _make_paths(
        self, train_path, val_path, test_path, wnids_path, words_path
    ):
        self.ids = []
        with open(wnids_path, "r") as idf:
            for nid in idf:
                nid = nid.strip()
                self.ids.append(nid)
        self.nid_to_words = defaultdict(list)
        with open(words_path, "r") as wf:
            for line in wf:
                nid, labels = line.split("\t")
                labels = list(map(lambda x: x.strip(), labels.split(",")))
                self.nid_to_words[nid].extend(labels)

        self.paths = {
            "train": [],  # [img_path, id, nid, box]
            "val": [],  # [img_path, id, nid, box]
            "test": [],  # img_path
        }

        # Get the test paths
        self.paths["test"] = list(
            map(lambda x: os.path.join(test_path, x), os.listdir(test_path))
        )
        # Get the validation paths and labels
        with open(os.path.join(val_path, "val_annotations.txt")) as valf:
            for line in valf:
                fname, nid, x0, y0, x1, y1 = line.split()
                fname = os.path.join(val_path, "images", fname)
                bbox = int(x0), int(y0), int(x1), int(y1)
                label_id = self.ids.index(nid)
                self.paths["val"].append((fname, label_id, nid, bbox))

        # Get the training paths
        train_nids = os.listdir(train_path)
        for nid in train_nids:
            anno_path = os.path.join(train_path, nid, nid + "_boxes.txt")
            imgs_path = os.path.join(train_path, nid, "images")
            label_id = self.ids.index(nid)
            with open(anno_path, "r") as annof:
                for line in annof:
                    fname, x0, y0, x1, y1 = line.split()
                    fname = os.path.join(imgs_path, fname)
                    bbox = int(x0), int(y0), int(x1), int(y1)
                    self.paths["train"].append((fname, label_id, nid, bbox))


class TinyImageNet(Dataset):
    """From https://gist.github.com/z-a-f/b862013c0dc2b540cf96a123a6766e54."""

    def __init__(
        self,
        root: str,
        split="train",
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        max_samples=None,
    ):
        self.split = split
        self.label_idx = 1  # from [image, id, nid, box]
        self.transform = transform
        self.target_transform = target_transform
        self.root = root

        self.IMAGE_SHAPE = (64, 64, 3)

        self.samples = []
        self.label_data = []

        self.max_samples = max_samples
        self.samples_num = len(self.samples)

        if self.max_samples is not None:
            self.samples_num = min(self.max_samples, self.samples_num)
            self.samples = np.random.permutation(self.samples)[
                : self.samples_num
            ]

        self.samples = np.zeros(
            (self.samples_num,) + self.IMAGE_SHAPE, dtype=np.float32
        )

        self.make_dataset(self.root)

    def make_dataset(self, directory: str):
        tinp = TinyImageNetPaths(directory)
        self.samples = tinp.paths[self.split]
        self.samples_num = len(self.samples)

        self.label_data = np.zeros((self.samples_num,), dtype=np.int)
        for idx in range(self.samples_num):
            s = self.samples[idx]
            # TODO: check that it works
            img = Image.open(s[0])
            # img = mpimg.imread(s[0])
            img = self._add_channels(img)
            img = Image.fromarray(np.uint8(img))
            self.samples[idx] = img
            if self.split != "test":
                self.label_data[idx] = s[self.label_idx]

    def _add_channels(self, img):
        while len(img.shape) < 3:  # third axis is the channels
            img = np.expand_dims(img, axis=-1)
        while (img.shape[-1]) < 3:
            img = np.concatenate([img, img[:, :, -1:]], axis=-1)
        return img

    def __len__(self):
        return self.samples_num

    def __getitem__(self, index):
        sample = self.samples[index]
        target = None if self.split == "test" else self.label_data[index]

        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return sample, target



from torchvision import transforms
from torchvision.datasets import ImageFolder


class MixingSet(ImageFolder):
    r"""Adapted from https://github.com/andyzoujm/pixmix/."""

    def __init__(self, root: str, main_dataset: str = "cifar"):
        assert main_dataset in [
            "cifar",
            "imagenet",
        ], """main_dataset should be equal to "cifar" or "imagenet". """

        if main_dataset == "cifar":
            dataset_transform = [
                transforms.Resize(64),
                transforms.RandomCrop(32),
            ]
        elif main_dataset == "resnet":
            dataset_transform = [
                transforms.Resize(256),
                transforms.RandomCrop(224),
            ]

        super().__init__(root, transform=dataset_transform)



import torch
from torch.utils.data import Dataset


class AggregatedDataset(Dataset):
    def __init__(self, dataset: Dataset, n_dataloaders: int) -> None:
        super().__init__()
        self.dataset = dataset
        self.n_dataloaders = n_dataloaders
        self.dataset_size = len(self.dataset)
        self.offset = self.dataset_size // self.n_dataloaders

    def __getitem__(self, idx: int):
        inputs, targets = zip(
            *[
                self.dataset[(idx + i * self.offset) % self.dataset_size]
                for i in range(self.n_dataloaders)
            ]
        )
        inputs = torch.cat(inputs, dim=0)
        targets = torch.as_tensor(targets)
        return inputs, targets

    def __len__(self):
        return self.dataset_size



# fmt: off
import math
from typing import Optional, Tuple, Union

import torch
import torch.nn as nn
from torch.nn.common_types import _size_2_t
from torch.nn.modules.utils import _pair


# fmt: on
class BatchLinear(nn.Module):
    r"""Applies a linear transformation using BatchEnsemble method to the
    incoming data: :math:`y=(x\circ \hat{R})W^{T}\circ \hat{S} + \hat{b}`.

    Source:
        Introduced by the paper `BatchEnsemble: An Alternative Approach to
        Efficient Ensemble and Lifelong Learning
        <https://arxiv.org/abs/2002.06715>`_, we present here an implementation
        of a Linear BatchEnsemble layer in `PyTorch <https://pytorch.org>`_
        heavily inspired by its `official implementation
        <https://github.com/google/edward2>`_ in `TensorFlow
        <https://www.tensorflow.org>`_.

    Args:
        in_features (int): size of each input sample.
        out_features (int): size of each output sample.
        num_estimators (int): number of estimators in the ensemble referred as
            :math:`M` here.
        bias (bool, optional): if ``True``, adds a learnable bias to the
            output. Defaults to ``True``.

    Attributes:
        weight:   the learnable weights (:math:`W`) of shape
            :math:`(H_{out}, H_{in})` shared between the estimators. The values
            are initialized from :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})`,
            where :math:`k = \frac{1}{H_{in}}`.
        R: 	the learnable matrice of shape :math:`(M, H_{in})` where each row
            consist of the vector :math:`r_{i}` corresponding to the
            :math:`i^{th}` ensemble member. The values are initialized from
            :math:`\mathcal{N}(1.0, 0.5)`.
        S: 	the learnable matrice of shape :math:`(M, H_{out})` where each row
            consist of the vector :math:`s_{i}` corresponding to the
            :math:`i^{th}` ensemble member. The values are initialized from
            :math:`\mathcal{N}(1.0, 0.5)`.
        bias: 	the learnable bias (:math:`b`) of shape :math:`(M, H_{out})`
            where each row corresponds to the bias of the :math:`i^{th}`
            ensemble member. If :attr:`bias` is ``True``, the values are
            initialized from :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})` where
            :math:`k = \frac{1}{H_{in}}`.

    Shape:
        - Input: :math:`(N, H_{in})` where :math:`N` is the batch size and
          :math:`H_{in} = \text{in_features}`.
        - Output: :math:`(N, H_{out})` where
          :math:`H_{out} = \text{out_features}`.

    Warning:
        Make sure that :math:`M | N` when calling :func:`forward()`.

    Examples:
        >>> # With three estimators
        >>> m = LinearBE(20, 30, 3)
        >>> input = torch.randn(8, 20)
        >>> output = m(input)
        >>> print(output.size())
        torch.Size([8, 30])
    """

    __constants__ = ["in_features", "out_features", "num_estimators"]
    in_features: int
    out_features: int
    n_estimator: int
    weight: torch.Tensor
    R: torch.Tensor
    S: torch.Tensor
    bias: Optional[torch.Tensor]

    def __init__(
        self,
        in_features: int,
        out_features: int,
        num_estimators: int,
        bias: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.num_estimators = num_estimators

        self.linear = nn.Linear(
            in_features=in_features, out_features=out_features, bias=False
        )

        self.R = nn.Parameter(
            torch.empty((num_estimators, in_features), **factory_kwargs)
        )
        self.S = nn.Parameter(
            torch.empty((num_estimators, out_features), **factory_kwargs)
        )
        if bias:
            self.bias = nn.Parameter(
                torch.empty((num_estimators, out_features), **factory_kwargs)
            )
        else:
            self.register_parameter("bias", None)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        # Setting a=sqrt(5) in kaiming_uniform is the same as initializing with
        # uniform(-1/sqrt(in_features), 1/sqrt(in_features)). For details, see
        # https://github.com/pytorch/pytorch/issues/57109
        # nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        nn.init.normal_(self.R, mean=1.0, std=0.5)
        nn.init.normal_(self.S, mean=1.0, std=0.5)
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(
                self.linear.weight
            )
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        batch_size = input.size(0)
        examples_per_estimator = torch.tensor(
            batch_size // self.num_estimators, device=input.device
        )
        extra = batch_size % self.num_estimators

        R = torch.repeat_interleave(self.R, examples_per_estimator, dim=0)
        R = torch.cat([R, R[:extra]], dim=0)  # .unsqueeze(-1).unsqueeze(-1)
        S = torch.repeat_interleave(self.S, examples_per_estimator, dim=0)
        S = torch.cat([S, S[:extra]], dim=0)  # .unsqueeze(-1).unsqueeze(-1)
        bias: Optional[torch.Tensor]
        if self.bias is not None:
            bias = torch.repeat_interleave(
                self.bias,
                examples_per_estimator,
                dim=0,
            )
            bias = torch.cat([bias, bias[:extra]], dim=0)
        else:
            bias = None

        return self.linear(input * R) * S + (bias if bias is not None else 0)

    def extra_repr(self) -> str:
        return (
            f"in_features={ self.in_features},"
            + f" out_features={self.out_features},"
            + f" num_estimators={self.num_estimators},"
            + f" bias={self.bias is not None}"
        )


class BatchConv2d(nn.Module):
    r"""Applies a 2D convolution over an input signal composed of several input
    planes using BatchEnsemble method to the incoming data.

    In the simplest case, the output value of the layer with input size
    :math:`(N, C_{in}, H_{in}, W_{in})` and output
    :math:`(N, C_{out}, H_{out}, W_{out})` can be precisely described as:

    .. math::
        \text{out}(N_i, C_{\text{out}_j})=\
        &\hat{b}(N_i,C_{\text{out}_j})
        +\hat{S}(N_{i},C_{\text{out}_j}) \\
        &\times \sum_{k = 0}^{C_{\text{in}} - 1}
        \text{weight}(C_{\text{out}_j}, k)\star (\text{input}(N_i, k)
        \times \hat{R}(N_i, k))

    Source:
        Introduced by the paper `BatchEnsemble: An Alternative Approach to
        Efficient Ensemble and Lifelong Learning
        <https://arxiv.org/abs/2002.06715>`_, we present here an implementation
        of a Conv2d BatchEnsemble layer in `PyTorch <https://pytorch.org>`_
        heavily inspired by its `official implementation
        <https://github.com/google/edward2>`_ in `TensorFlow
        <https://www.tensorflow.org>`_.

    Args:
        in_channels (int): number of channels in the input image.
        out_channels (int): number of channels produced by the convolution.
        kernel_size (int or tuple): size of the convolving kernel.
        num_estimators (int): number of estimators in the ensemble referred as
            :math:`M` here.
        stride (int or tuple, optional): stride of the convolution. Defaults to
            ``1``.
        padding (int, tuple or str, optional): padding added to all four sides
            of the input. Defaults to ``0``.
        dilation (int or tuple, optional): spacing between kernel elements.
            Defaults to ``1``.
        groups (int, optional): number of blocked connections from input
            channels to output channels. Defaults to ``1``.
        bias (bool, optional): if ``True``, adds a learnable bias to the
            output. Defaults to ``True``.

    Attributes:
        weight:     the learnable weights of the module of shape
            :math:`(\text{out_channels}, \frac{\text{in_channels}}
            {\text{groups}},`:math:`\text{kernel_size[0]},
            \text{kernel_size[1]})` shared between the estimators. The values
            of these weights are sampled from :math:`\mathcal{U}(-\sqrt{k},
            \sqrt{k})` where :math:`k = \frac{\text{groups}}{C_\text{in} *
            \prod_{i=0}^{1}\text{kernel_size}[i]}`.
        R: 	the learnable matrice of shape :math:`(M, C_{in})` where each row
            consist of the vector :math:`r_{i}` corresponding to the
            :math:`i^{th}` ensemble member. The values are initialized from
            :math:`\mathcal{N}(1.0, 0.5)`.
        S: 	the learnable matrice of shape :math:`(M, C_{out})` where each row
            consist of the vector :math:`s_{i}` corresponding to the
            :math:`i^{th}` ensemble member. The values are initialized from
            :math:`\mathcal{N}(1.0, 0.5)`.
        bias: 	the learnable bias (:math:`b`) of shape :math:`(M, C_{out})`
            where each row corresponds to the bias of the :math:`i^{th}`
            ensemble member. If :attr:`bias` is ``True``, the values are
            initialized from :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})` where
            :math:`k=\frac{\text{groups}}{C_\text{in}*\prod_{i=0}^{1}
            \text{kernel_size}[i]}`.

    Shape:
        - Input: :math:`(N, C_{in}, H_{in}, W_{in})`.
        - Output: :math:`(N, C_{out}, H_{out}, W_{out})`.

        .. math::
            H_{out} = \left\lfloor\frac{H_{in} + 2 \times \text{padding}[0] -
            \text{dilation}[0] \times (\text{kernel_size}[0] - 1) - 1}
            {\text{stride}[0]} + 1\right\rfloor

        .. math::
            W_{out} = \left\lfloor\frac{W_{in}  + 2 \times \text{padding}[1] -
            \text{dilation}[1] \times (\text{kernel_size}[1] - 1) - 1}
            {\text{stride}[1]} + 1\right\rfloor

    .. warning::

        Make sure that :math:`M | N` when calling :func:`forward()`.

    Examples:

        >>> # With square kernels, four estimators and equal stride
        >>> m = Conv2dBE(3, 32, 3, 4, stride=1)
        >>> input = torch.randn(8, 3, 16, 16)
        >>> output = m(input)
        >>> print(output.size())
        torch.Size([8, 32, 14, 14])
    """

    __constants__ = [
        "stride",
        "padding",
        "dilation",
        "groups",
        "in_channels",
        "out_channels",
        "kernel_size",
        "num_estimators",
    ]
    in_channels: int
    out_channels: int
    kernel_size: Tuple[int, ...]
    n_estimator: int
    stride: Tuple[int, ...]
    padding: Union[str, Tuple[int, ...]]
    dilation: Tuple[int, ...]
    groups: int
    weight: torch.Tensor
    R: torch.Tensor
    S: torch.Tensor
    bias: Optional[torch.Tensor]

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_2_t,
        num_estimators: int,
        stride: _size_2_t = 1,
        padding: Union[str, _size_2_t] = 0,
        dilation: _size_2_t = 1,
        groups: int = 1,
        bias: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = _pair(kernel_size)
        self.num_estimators = num_estimators
        self.stride = _pair(stride)
        self.padding = padding if isinstance(padding, str) else _pair(padding)
        self.dilation = _pair(dilation)
        self.groups = groups

        self.conv = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=False,
        )
        self.R = nn.Parameter(
            torch.empty((num_estimators, in_channels), **factory_kwargs)
        )
        self.S = nn.Parameter(
            torch.empty((num_estimators, out_channels), **factory_kwargs)
        )
        if bias:
            self.bias = nn.Parameter(
                torch.empty((num_estimators, out_channels), **factory_kwargs)
            )
        else:
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        # Setting a=sqrt(5) in kaiming_uniform is the same as initializing with
        # uniform(-1/sqrt(k), 1/sqrt(k)), where
        # k = weight.size(1) * prod(*kernel_size)
        # For more details see:
        # https://github.com/pytorch/pytorch/issues/15314#issuecomment-477448573
        # nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        nn.init.normal_(self.R, mean=1.0, std=0.5)
        nn.init.normal_(self.S, mean=1.0, std=0.5)
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.conv.weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        batch_size = input.size(0)
        examples_per_estimator = batch_size // self.num_estimators
        extra = batch_size % self.num_estimators

        R = (
            torch.repeat_interleave(
                self.R,
                torch.full(
                    [self.num_estimators],
                    examples_per_estimator,
                    device=self.R.device,
                ),
                dim=0,
            )
            .unsqueeze(-1)
            .unsqueeze(-1)
        )
        R = torch.cat([R, R[:extra]], dim=0)  # .unsqueeze(-1).unsqueeze(-1)
        S = (
            torch.repeat_interleave(
                self.S,
                torch.full(
                    [self.num_estimators],
                    examples_per_estimator,
                    device=self.S.device,
                ),
                dim=0,
            )
            .unsqueeze(-1)
            .unsqueeze(-1)
        )
        S = torch.cat([S, S[:extra]], dim=0)  #

        bias: Optional[torch.Tensor]
        if self.bias is not None:
            bias = (
                torch.repeat_interleave(
                    self.bias,
                    torch.full(
                        [self.num_estimators],
                        examples_per_estimator,
                        device=self.bias.device,
                    ),
                    dim=0,
                )
                .unsqueeze(-1)
                .unsqueeze(-1)
            )

            bias = torch.cat([bias, bias[:extra]], dim=0)
        else:
            bias = None

        return self.conv(input * R) * S + (bias if bias is not None else 0)

    def extra_repr(self):
        s = (
            "{in_channels}, {out_channels}, kernel_size={kernel_size}"
            ", num_estimators={num_estimators}, stride={stride}"
        )
        return s.format(**self.__dict__)



# fmt: off
from argparse import ArgumentParser
from typing import Any, Dict, Literal

import torch
import torch.nn as nn

from torch_uncertainty.models.resnet import (
    packed_resnet18,
    packed_resnet34,
    packed_resnet50,
    packed_resnet101,
    packed_resnet152,
)
from torch_uncertainty.routines.classification import ClassificationEnsemble

# fmt: on
archs = [
    packed_resnet18,
    packed_resnet34,
    packed_resnet50,
    packed_resnet101,
    packed_resnet152,
]
choices = [18, 34, 50, 101, 152]


class PackedResNet(ClassificationEnsemble):
    r"""LightningModule for Packed-Ensembles ResNet.

    Args:
        num_classes (int): Number of classes to predict.
        num_estimators (int): Number of estimators in the ensemble.
        in_channels (int): Number of input channels.
        alpha (int): Expansion factor affecting the width of the estimators.
        gamma (int): Number of groups within each estimator.
        arch (int):
            Determines which ResNet architecture to use:

            - ``18``: ResNet-18
            - ``32``: ResNet-32
            - ``50``: ResNet-50
            - ``101``: ResNet-101
            - ``152``: ResNet-152

        loss (torch.nn.Module): Training loss.
        optimization_procedure (Any): Optimization procedure, corresponds to
            what expect the `LightningModule.configure_optimizers()
            <https://pytorch-lightning.readthedocs.io/en/stable/common/lightning_module.html#configure-optimizers>`_
            method.
        use_entropy (bool, optional): Indicates whether to use the entropy
            values as the OOD criterion or not. Defaults to ``False``.
        use_logits (bool, optional): Indicates whether to use the logits as the
            OOD criterion or not. Defaults to ``False``.
        use_mi (bool, optional): Indicates whether to use the mutual
            information as the OOD criterion or not. Defaults to ``False``.
        use_variation_ratio (bool, optional): Indicates whether to use the
            variation ratio as the OOD criterion or not. Defaults to ``False``.

    Note:

        The OOD criterion is by defaults the confidence score.

    Warning:

        Make sure at most only one of :attr:`use_entropy`, :attr:`use_logits`,
        :attr:`use_mi` and :attr:`use_variation_ratio` attributes is set to
        ``True``. Otherwise a :class:`ValueError()` will be raised.

    Raises:
            ValueError: If :attr:`alpha`:math:`<1`.
            ValueError: If :attr:`gamma`:math:`<1`.
    """

    def __init__(
        self,
        num_classes: int,
        num_estimators: int,
        in_channels: int,
        alpha: int,
        gamma: int,
        arch: Literal[18, 34, 50, 101, 152],
        loss: nn.Module,
        optimization_procedure: Any,
        use_entropy: bool = False,
        use_logits: bool = False,
        use_mi: bool = False,
        use_variation_ratio: bool = False,
        **kwargs: Dict[str, Any],
    ) -> None:
        super().__init__(
            num_classes=num_classes,
            num_estimators=num_estimators,
            use_entropy=use_entropy,
            use_logits=use_logits,
            use_mi=use_mi,
            use_variation_ratio=use_variation_ratio,
        )

        if alpha <= 0:
            raise ValueError(f"Attribute `alpha` should be > 0, not {alpha}")
        if gamma < 1:
            raise ValueError(f"Attribute `gamma` should be >= 1, not {gamma}")

        # construct config
        self.save_hyperparameters(ignore=["loss", "optimization_procedure"])

        self.loss = loss
        self.optimization_procedure = optimization_procedure

        self.model = archs[choices.index(arch)](
            in_channels=in_channels,
            num_estimators=num_estimators,
            alpha=alpha,
            gamma=gamma,
            num_classes=num_classes,
        )

        # to log the graph
        self.example_input_array = torch.randn(1, in_channels, 32, 32)

    def configure_optimizers(self) -> dict:
        return self.optimization_procedure(self)

    @property
    def criterion(self) -> nn.Module:
        return self.loss()

    def forward(self, input: torch.Tensor) -> torch.Tensor:  # type: ignore
        input = input.repeat(1, self.num_estimators, 1, 1)
        return self.model.forward(input)

    @staticmethod
    def add_model_specific_args(
        parent_parser: ArgumentParser,
    ) -> ArgumentParser:
        """Defines the model's attributes via command-line options:

        - ``--arch [int]``: defines :attr:`arch`. Defaults to ``18``.
        - ``--num_estimators [int]``: defines :attr:`num_estimators`. Defaults
          to ``1``.
        - ``--alpha [int]``: defines :attr:`alpha`. Defaults to ``1``.
        - ``--gamma [int]``: defines :attr:`gamma`. Defaults to ``1``.
        - ``--entropy``: sets :attr:`use_entropy` to ``True``.
        - ``--logits``: sets :attr:`use_logits` to ``True``.
        - ``--mutual_information``: sets :attr:`use_mi` to ``True``.
        - ``--variation_ratio``: sets :attr:`use_variation_ratio` to ``True``.

        Example:

            .. parsed-literal::

                python script.py --arch 18 --num_estimators 4 --alpha 2
        """
        parent_parser.add_argument(
            "--arch",
            type=int,
            default=18,
            choices=choices,
            help="Type of ResNet",
        )
        parent_parser.add_argument("--alpha", type=int, default=2)
        parent_parser.add_argument(
            "--entropy", dest="use_entropy", action="store_true"
        )
        parent_parser.add_argument("--gamma", type=int, default=1)
        parent_parser.add_argument(
            "--logits", dest="use_logits", action="store_true"
        )
        parent_parser.add_argument(
            "--mutual_information", dest="uses_mi", action="store_true"
        )
        parent_parser.add_argument("--num_estimators", type=int, default=4)
        parent_parser.add_argument(
            "--variation_ratio", dest="use_variation_ratio", action="store_true"
        )
        return parent_parser



# fmt: off
from typing import Any, Union

import torch.nn as nn
from einops import rearrange
from torch import Tensor
from torch.nn.common_types import _size_2_t


# fmt: on
class PackedLinear(nn.Module):
    r"""Packed-Ensembles-style Linear layer.

    This layer computes fully-connected operation for a given number of
    estimators (:attr:`num_estimators`) using a `1x1` convolution.

    Args:
        in_features (int): Number of input features of the linear layer.
        out_features (int): Number of channels produced by the linear layer.
        num_estimators (int): The number of estimators grouped in the layer.
        bias (bool, optional): It ``True``, adds a learnable bias to the
            output. Defaults to ``True``.
        groups (int, optional): Number of blocked connections from input
            channels to output channels. Defaults to ``1``.
        rearrange (bool, optional): Rearrange the input and outputs for
            compatibility with previous and later layers. Defaults to ``True``.

        Note:
            Each ensemble member will only see
            :math:`\frac{\text{in\_features}}{\text{n\_estimators}}` features,
            so when using :attr:`groups` you should make sure that
            :attr:`in_features` and :attr:`out_features` are both divisible by
            :attr:`n_estimators` :math:`\times`:attr:`groups`. However, the
            number of input and output features will be changed to comply with
            this constraint.

        Note:
            The input should be of size (`batch_size`, :attr:`in_features`, 1,
            1). The (often) necessary rearrange operation is executed by
            default.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        num_estimators: int,
        bias: bool = True,
        groups: int = 1,
        rearrange: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()

        self.num_estimators = num_estimators
        self.rearrange = rearrange

        # fix if not divisible by groups
        if in_features % (num_estimators * groups):
            in_features += num_estimators - in_features % (
                num_estimators * groups
            )
        if out_features % (num_estimators * groups):
            out_features += num_estimators - out_features % (
                num_estimators * groups
            )

        self.conv1x1 = nn.Conv2d(
            in_channels=in_features,
            out_channels=out_features,
            kernel_size=1,
            stride=1,
            padding=0,
            dilation=1,
            groups=num_estimators * groups,
            bias=bias,
            padding_mode="zeros",
            **factory_kwargs,
        )

    def _rearrange_forward(self, x: Tensor) -> Tensor:
        x = x.unsqueeze(-1).unsqueeze(-1)
        x = rearrange(x, "(m e) c h w -> e (m c) h w", m=self.num_estimators)
        x = self.conv1x1(x)
        x = rearrange(x, "e (m c) h w -> (m e) c h w", m=self.num_estimators)
        return x.squeeze(-1).squeeze(-1)

    def forward(self, input: Tensor) -> Tensor:
        if self.rearrange:
            return self._rearrange_forward(input)
        else:
            return self.conv1x1(input)


class PackedConv2d(nn.Module):
    r"""Packed-Ensembles-style Conv2d layer.

    Args:
        in_channels (int): Number of channels in the input image.
        out_channels (int): Number of channels produced by the convolution.
        kernel_size (int or tuple): Size of the convolving kernel.
        num_estimators (int): Number of estimators in the ensemble.
        stride (int or tuple, optional): Stride of the convolution.
            Defaults to ``1``.
        padding (int, tuple or str, optional): Padding added to all four sides
            of the input. Defaults to ``0``.
        dilation (int or tuple, optional): Spacing between kernel elements.
            Defaults to ``1``.
        groups (int, optional): Number of blocked connexions from input
            channels to output channels for each estimator. Defaults to ``1``.
        minimum_channels_per_group (int, optional): Smallest possible number of
            hannels per group.
        bias (bool, optional): If ``True``, adds a learnable bias to the
            output. Defaults to ``True``.

    Note:
        Each ensemble member will only see
        :math:`\frac{\text{in\_channels}}{\text{num\_estimators}}` channels,
        so when using :attr:`groups` you should make sure that
        :attr:`in_channels` and :attr:`out_channels` are both divisible by
        :attr:`num_estimators` :math:`\times`:attr:`groups`. However, the
        number of input and output channels will be changed to comply with this
        constraint.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_2_t,
        num_estimators: int,
        stride: _size_2_t = 1,
        padding: Union[str, _size_2_t] = 0,
        dilation: _size_2_t = 1,
        groups: int = 1,
        minimum_channels_per_group: int = 64,
        bias: bool = True,
        device: Union[Any, None] = None,
        dtype: Union[Any, None] = None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()

        self.num_estimators = num_estimators

        while (
            in_channels % (num_estimators * groups) != 0
            or in_channels // (num_estimators * groups)
            < minimum_channels_per_group
        ) and groups > 1:
            groups -= 1

        # fix if not divisible by groups
        if in_channels % (num_estimators * groups):
            in_channels += (
                num_estimators - in_channels % num_estimators * groups
            )
        if out_channels % (num_estimators * groups):
            out_channels += (
                num_estimators - out_channels % num_estimators * groups
            )

        self.conv = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=num_estimators * groups,
            bias=bias,
            padding_mode="zeros",
            **factory_kwargs,
        )

    def forward(self, input: Tensor) -> Tensor:
        return self.conv(input)



from pathlib import Path, PurePath
from typing import Any, Callable, Optional, Tuple

from torchvision.datasets import ImageFolder, ImageNet


class ImageNetR(ImageFolder):
    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        is_valid_file: Optional[Callable[[str], bool]] = None,
        split: str = None,
    ):
        super().__init__(
            root=root / Path("imagenet-r"),
            transform=transform,
            target_transform=target_transform,
        )
        imagenet = ImageNet(root, split="val")
        self.wnid_to_idx = imagenet.wnid_to_idx
        for id, sample in enumerate(self.samples):
            cls = PurePath(sample[0]).parts[-2]
            self.samples[id] = (sample[0], self.wnid_to_idx[cls])

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        path, target = self.samples[index]
        return super().__getitem__(index)

    def __len__(self) -> int:
        return len(self.samples)



# fmt: off
from typing import List, Type, Union

import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

# fmt: on
__all__ = [
    "resnet18",
    "resnet34",
    "resnet50",
    "resnet101",
    "resnet152",
]


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(
        self,
        in_planes: int,
        planes: int,
        stride: int = 1,
        groups: int = 1,
    ):
        super(BasicBlock, self).__init__()

        self.conv1 = nn.Conv2d(
            in_planes,
            planes,
            kernel_size=3,
            stride=stride,
            padding=1,
            groups=groups,
            bias=False,
        )
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(
            planes,
            planes,
            kernel_size=3,
            stride=1,
            padding=1,
            groups=groups,
            bias=False,
        )
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(
                    in_planes,
                    self.expansion * planes,
                    kernel_size=1,
                    stride=stride,
                    groups=groups,
                    bias=False,
                ),
                nn.BatchNorm2d(self.expansion * planes),
            )

    def forward(self, x: Tensor) -> Tensor:
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, in_planes, planes, stride=1, groups=1):
        super(Bottleneck, self).__init__()

        self.conv1 = nn.Conv2d(
            in_planes,
            planes,
            kernel_size=1,
            groups=groups,
            bias=False,
        )
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(
            planes,
            planes,
            kernel_size=3,
            stride=stride,
            padding=1,
            groups=groups,
            bias=False,
        )
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(
            planes,
            self.expansion * planes,
            kernel_size=1,
            groups=groups,
            bias=False,
        )
        self.bn3 = nn.BatchNorm2d(self.expansion * planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(
                    in_planes,
                    self.expansion * planes,
                    kernel_size=1,
                    stride=stride,
                    groups=groups,
                    bias=False,
                ),
                nn.BatchNorm2d(self.expansion * planes),
            )

    def forward(self, x: Tensor) -> Tensor:
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class _ResNet(nn.Module):
    def __init__(
        self,
        block: Type[Union[BasicBlock, Bottleneck]],
        num_blocks: List[int],
        in_channels: int,
        num_classes: int,
        groups: int,
        # dataset: str = "cifar",
    ) -> None:
        super().__init__()
        # assert dataset in [
        #     "cifar",
        #     "mnist",
        #     "tinyimagenet",
        #     "imagenet",
        # ], "The dataset is not taken in charge by this implementation."
        # self.dataset = dataset
        self.in_planes = 64
        block_planes = self.in_planes

        # if self.dataset == "imagenet":
        #     self.conv1 = nn.Conv2d(
        #         3 * self.num_estimators,
        #         block_planes,
        #         kernel_size=7,
        #         stride=2,
        #         padding=3,
        #         groups=1,
        #         num_estimators=num_estimators,
        #         bias=False,
        #     )
        # elif self.dataset == "mnist":
        #     self.conv1 = nn.Conv2d(
        #         1 * self.num_estimators,
        #         block_planes,
        #         kernel_size=3,
        #         stride=1,
        #         padding=1,
        #         groups=1,
        #         num_estimators=num_estimators,
        #         bias=False,
        #     )
        # else:

        # No groups in the first layer
        self.conv1 = nn.Conv2d(
            in_channels,
            block_planes,
            kernel_size=3,
            stride=1,
            padding=1,
            groups=1,
            bias=False,
        )

        self.bn1 = nn.BatchNorm2d(block_planes)

        self.optional_pool = nn.Identity()

        self.layer1 = self._make_layer(
            block,
            block_planes,
            num_blocks[0],
            stride=1,
            groups=groups,
        )
        self.layer2 = self._make_layer(
            block,
            block_planes * 2,
            num_blocks[1],
            stride=2,
            groups=groups,
        )
        self.layer3 = self._make_layer(
            block,
            block_planes * 4,
            num_blocks[2],
            stride=2,
            groups=groups,
        )
        self.layer4 = self._make_layer(
            block,
            block_planes * 8,
            num_blocks[3],
            stride=2,
            groups=groups,
        )

        self.pool = nn.AdaptiveAvgPool2d(output_size=1)
        self.flatten = nn.Flatten(1)

        self.linear = nn.Linear(
            block_planes * 8 * block.expansion,
            num_classes,
        )

    def _make_layer(
        self,
        block: Union[Type[BasicBlock], Type[Bottleneck]],
        planes: int,
        num_blocks: int,
        stride: int,
        groups: int,
    ) -> nn.Module:
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(
                block(
                    in_planes=self.in_planes,
                    planes=planes,
                    stride=stride,
                    groups=groups,
                )
            )
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.optional_pool(out)
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = self.pool(out)
        out = self.flatten(out)
        out = self.linear(out)
        return out


def resnet18(
    in_channels: int,
    num_classes: int,
    groups: int,
) -> _ResNet:
    """ResNet-18 from `Deep Residual Learning for Image Recognition
    <https://arxiv.org/pdf/1512.03385.pdf>`_.

    Args:
        in_channels (int): Number of input channels.
        num_classes (int): Number of classes to predict.
        groups (int): Number of groups in convolutions.

    Returns:
        _PackedResNet: A ResNet-18.
    """
    return _ResNet(
        block=BasicBlock,
        num_blocks=[2, 2, 2, 2],
        in_channels=in_channels,
        num_classes=num_classes,
        groups=groups,
    )


def resnet34(
    in_channels: int,
    num_classes: int,
    groups: int,
) -> _ResNet:
    """ResNet-34 from `Deep Residual Learning for Image Recognition
    <https://arxiv.org/pdf/1512.03385.pdf>`_.

    Args:
        in_channels (int): Number of input channels.
        num_classes (int): Number of classes to predict.
        groups (int): Number of groups in convolutions.

    Returns:
        _PackedResNet: A ResNet-34.
    """
    return _ResNet(
        block=BasicBlock,
        num_blocks=[3, 4, 6, 3],
        in_channels=in_channels,
        groups=groups,
        num_classes=num_classes,
    )


def resnet50(
    in_channels: int,
    num_classes: int,
    groups: int,
) -> _ResNet:
    """ResNet-50 from `Deep Residual Learning for Image Recognition
    <https://arxiv.org/pdf/1512.03385.pdf>`_.

    Args:
        in_channels (int): Number of input channels.
        num_classes (int): Number of classes to predict.
        groups (int): Number of groups in convolutions.

    Returns:
        _PackedResNet: A ResNet-50.
    """
    return _ResNet(
        block=Bottleneck,
        num_blocks=[3, 4, 6, 3],
        in_channels=in_channels,
        groups=groups,
        num_classes=num_classes,
    )


def resnet101(
    in_channels: int,
    num_classes: int,
    groups: int,
) -> _ResNet:
    """ResNet-101 from `Deep Residual Learning for Image Recognition
    <https://arxiv.org/pdf/1512.03385.pdf>`_.

    Args:
        in_channels (int): Number of input channels.
        num_classes (int): Number of classes to predict.
        groups (int): Number of groups in convolutions.

    Returns:
        _PackedResNet: A ResNet-101.
    """
    return _ResNet(
        block=Bottleneck,
        num_blocks=[3, 4, 23, 3],
        in_channels=in_channels,
        groups=groups,
        num_classes=num_classes,
    )


def resnet152(
    in_channels: int,
    num_classes: int,
    groups: int,
) -> _ResNet:
    """ResNet-152 from `Deep Residual Learning for Image Recognition
    <https://arxiv.org/pdf/1512.03385.pdf>`_.

    Args:
        in_channels (int): Number of input channels.
        num_classes (int): Number of classes to predict.
        groups (int): Number of groups in convolutions.

    Returns:
        _PackedResNet: A ResNet-152.
    """
    return _ResNet(
        block=Bottleneck,
        num_blocks=[3, 8, 36, 3],
        in_channels=in_channels,
        groups=groups,
        num_classes=num_classes,
    )



# fmt: off
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Callable, Type, Union

import pytorch_lightning as pl
from pytorch_lightning.callbacks import LearningRateMonitor
from pytorch_lightning.callbacks.early_stopping import EarlyStopping
from pytorch_lightning.callbacks.model_checkpoint import ModelCheckpoint
from pytorch_lightning.loggers.tensorboard import TensorBoardLogger
from torch.nn import Module
from torchinfo import summary

import numpy as np

from .routines.classification import ClassificationSingle
from .utils import get_version

# fmt: on


def main(
    network: Type[ClassificationSingle],
    datamodule: Type[pl.LightningDataModule],
    loss: Module,
    optimization_procedure: Callable[[Module], dict],
    root: Union[Path, str],
    net_name: str,
    args: Namespace,
) -> None:
    if args.seed:
        pl.seed_everything(args.seed, workers=True)

    if isinstance(root, str):
        root = Path(root)

    # datamodule
    args.root = str(root / "data")
    dm = datamodule(**vars(args))

    # model
    model = network(
        loss=loss,
        optimization_procedure=optimization_procedure,
        num_classes=dm.num_classes,
        in_channels=dm.num_channels,
        **vars(args),
    )

    # logger
    tb_logger = TensorBoardLogger(
        str(root / "logs"),
        name=net_name,
        default_hp_metric=False,
        log_graph=args.log_graph,
        version=args.test,
    )

    # callbacks
    save_checkpoints = ModelCheckpoint(
        monitor="hp/val_acc",
        mode="max",
        save_last=True,
        save_weights_only=True,
    )

    # Select the best model, monitor the lr and stop if NaN
    callbacks = [
        save_checkpoints,
        LearningRateMonitor(logging_interval="step"),
        EarlyStopping(monitor="hp/val_nll", patience=np.inf, check_finite=True),
    ]

    # trainer
    trainer = pl.Trainer.from_argparse_args(
        args,
        callbacks=callbacks,
        logger=tb_logger,
        deterministic=(args.seed is not None),
    )

    if args.summary:
        summary(model, input_size=model.example_input_array.shape)
    elif args.test is not None:
        ckpt_file, _ = get_version(
            root=(root / "logs" / net_name), version=args.test
        )
        trainer.test(model, datamodule=dm, ckpt_path=str(ckpt_file))
    else:
        # training and testing
        trainer.fit(model, dm)
        trainer.test(datamodule=dm, ckpt_path="best")


def cli_main(
    network: Type[ClassificationSingle],
    datamodule: Type[pl.LightningDataModule],
    loss: Module,
    optimization_procedure: Callable[[Module], dict],
    root: Union[Path, str],
    net_name: str,
) -> None:
    parser = ArgumentParser("torch-uncertainty")
    parser.add_argument("--seed", type=int, default=None)
    parser.add_argument("--test", type=int, default=None)
    parser.add_argument("--summary", dest="summary", action="store_true")
    parser.add_argument("--log_graph", dest="log_graph", action="store_true")

    parser = pl.Trainer.add_argparse_args(parser)
    parser = datamodule.add_argparse_args(parser)
    parser = network.add_model_specific_args(parser)
    args = parser.parse_args()
    main(
        network, datamodule, loss, optimization_procedure, root, net_name, args
    )



# flake8: noqa
from .cifar10 import CIFAR10DataModule
from .cifar100 import CIFAR100DataModule
from .imagenet import ImageNetDataModule
from .mnist import MNISTDataModule
from .tiny_imagenet import TinyImageNetDataModule



# fmt: off
from argparse import Namespace
from typing import List, Tuple, Union

import pytorch_lightning as pl
import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange
from pytorch_lightning.utilities.memory import get_model_size_mb
from pytorch_lightning.utilities.types import EPOCH_OUTPUT, STEP_OUTPUT
from torchmetrics import (
    AUROC,
    Accuracy,
    AveragePrecision,
    CalibrationError,
    MetricCollection,
)

from ..metrics import (
    FPR95,
    Disagreement,
    Entropy,
    MutualInformation,
    NegativeLogLikelihood,
    VariationRatio,
)


# fmt:on
class ClassificationSingle(pl.LightningModule):
    def __init__(
        self,
        num_classes: int,
        use_entropy: bool = False,
        use_logits: bool = False,
    ) -> None:
        super().__init__()

        # FIXME: use ValueError instead
        assert (
            use_logits + use_entropy
        ) <= 1, "You cannot choose more than one OOD criterion."

        self.num_classes = num_classes
        self.use_logits = use_logits
        self.use_entropy = use_entropy

        # metrics
        cls_metrics = MetricCollection(
            {
                "nll": NegativeLogLikelihood(),
                "acc": Accuracy(
                    task="multiclass", num_classes=self.num_classes
                ),
                "ece": CalibrationError(
                    task="multiclass", num_classes=self.num_classes
                ),
            },
            compute_groups=False,
        )

        ood_metrics = MetricCollection(
            {
                "fpr95": FPR95(pos_label=1),
                "auroc": AUROC(task="binary"),
                "aupr": AveragePrecision(task="binary"),
            },
            compute_groups=[["auroc", "aupr"], ["fpr95"]],
        )

        self.val_metrics = cls_metrics.clone(prefix="hp/val_")
        self.test_cls_metrics = cls_metrics.clone(prefix="hp/test_")
        self.test_ood_metrics = ood_metrics.clone(prefix="hp/test_")

        self.test_entropy_id = Entropy()
        self.test_entropy_ood = Entropy()

    @property
    def criterion(self) -> nn.Module:
        raise NotImplementedError()

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    def on_train_start(self) -> None:
        # hyperparameters for performances
        param = {}
        param["storage"] = f"{get_model_size_mb(self)} MB"
        if self.logger is not None:
            self.logger.log_hyperparams(
                Namespace(**param),
                {
                    "hp/val_nll": 0,
                    "hp/val_acc": 0,
                    "hp/test_acc": 0,
                    "hp/test_nll": 0,
                    "hp/test_ece": 0,
                    "hp/test_entropy_id": 0,
                    "hp/test_entropy_ood": 0,
                    "hp/test_aupr": 0,
                    "hp/test_auroc": 0,
                    "hp/test_fpr95": 0,
                },
            )

    def training_step(
        self, batch: Tuple[torch.Tensor, torch.Tensor], batch_idx: int
    ) -> STEP_OUTPUT:
        inputs, targets = batch
        logits = self.forward(inputs)
        loss = self.criterion(logits, targets)
        self.log("train_loss", loss)
        return loss

    def validation_step(
        self, batch: Tuple[torch.Tensor, torch.Tensor], batch_idx: int
    ) -> None:
        inputs, targets = batch
        logits = self.forward(inputs)
        probs = F.softmax(logits, dim=-1)
        self.val_metrics.update(probs, targets)

    def validation_epoch_end(
        self, outputs: Union[EPOCH_OUTPUT, List[EPOCH_OUTPUT]]
    ) -> None:
        self.log_dict(self.val_metrics.compute())
        self.val_metrics.reset()

    def test_step(
        self,
        batch: Tuple[torch.Tensor, torch.Tensor],
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        inputs, targets = batch
        logits = self.forward(inputs)
        probs = F.softmax(logits, dim=-1)
        confs, _ = probs.max(dim=-1)

        if self.use_logits:
            ood_values, _ = -logits.max(dim=-1)
        elif self.use_entropy:
            ood_values = torch.special.entr(probs).sum(dim=-1)
        else:
            ood_values = -confs

        if dataloader_idx == 0:
            self.test_cls_metrics.update(probs, targets)
            self.test_ood_metrics.update(ood_values, torch.zeros_like(targets))
            self.test_entropy_id(probs)
            self.log(
                "hp/test_entropy_id",
                self.test_entropy_id,
                on_epoch=True,
                add_dataloader_idx=False,
            )
        else:
            self.test_ood_metrics.update(ood_values, torch.ones_like(targets))
            self.test_entropy_ood(probs)
            self.log(
                "hp/test_entropy_ood",
                self.test_entropy_ood,
                on_epoch=True,
                add_dataloader_idx=False,
            )

    def test_epoch_end(
        self, outputs: Union[EPOCH_OUTPUT, List[EPOCH_OUTPUT]]
    ) -> None:
        self.log_dict(
            self.test_cls_metrics.compute(),
        )
        self.log_dict(
            self.test_ood_metrics.compute(),
        )
        self.test_cls_metrics.reset()
        self.test_ood_metrics.reset()


class ClassificationEnsemble(ClassificationSingle):
    def __init__(
        self,
        num_classes: int,
        num_estimators: int,
        use_entropy: bool = False,
        use_logits: bool = False,
        use_mi: bool = False,
        use_variation_ratio: bool = False,
    ) -> None:
        super().__init__(
            num_classes=num_classes,
            use_entropy=use_entropy,
            use_logits=use_logits,
        )

        self.num_estimators = num_estimators

        self.use_mi = use_mi
        self.use_variation_ratio = use_variation_ratio

        # FIXME: use ValueError instead
        assert (
            self.use_logits
            + self.use_entropy
            + self.use_mi
            + self.use_variation_ratio
        ) <= 1, "You cannot choose more than one OOD criterion."

        # metrics for ensembles only
        ens_metrics = MetricCollection(
            {
                "disagreement": Disagreement(),
                "mi": MutualInformation(),
                "entropy": Entropy(),
            }
        )
        self.test_id_ens_metrics = ens_metrics.clone(prefix="hp/test_id_ens_")
        self.test_ood_ens_metrics = ens_metrics.clone(prefix="hp/test_ood_ens_")

    def on_train_start(self) -> None:
        # hyperparameters for performances
        param = {}
        param["storage"] = f"{get_model_size_mb(self)} MB"
        if self.logger is not None:
            self.logger.log_hyperparams(
                Namespace(**param),
                {
                    "hp/val_nll": 0,
                    "hp/val_acc": 0,
                    "hp/test_acc": 0,
                    "hp/test_nll": 0,
                    "hp/test_ece": 0,
                    "hp/test_entropy_id": 0,
                    "hp/test_entropy_ood": 0,
                    "hp/test_aupr": 0,
                    "hp/test_auroc": 0,
                    "hp/test_fpr95": 0,
                    "hp/test_id_ens_disagreement": 0,
                    "hp/test_id_ens_mi": 0,
                    "hp/test_id_ens_entropy": 0,
                    "hp/test_ood_ens_disagreement": 0,
                    "hp/test_ood_ens_mi": 0,
                    "hp/test_ood_ens_entropy": 0,
                },
            )

    def training_step(
        self, batch: Tuple[torch.Tensor, torch.Tensor], batch_idx: int
    ) -> STEP_OUTPUT:
        inputs, targets = batch
        targets = targets.repeat(self.num_estimators)
        return super().training_step((inputs, targets), batch_idx)

    def validation_step(  # type: ignore
        self, batch: Tuple[torch.Tensor, torch.Tensor], batch_idx: int
    ) -> None:
        inputs, targets = batch
        logits = self.forward(inputs)
        # logits = logits.reshape(self.num_estimators, -1, logits.size(-1))
        logits = rearrange(logits, "(n b) c -> b n c", n=self.num_estimators)
        probs_per_est = F.softmax(logits, dim=-1)
        probs = probs_per_est.mean(dim=1)
        self.val_metrics.update(probs, targets)

    def test_step(
        self,
        batch: Tuple[torch.Tensor, torch.Tensor],
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        inputs, targets = batch
        logits = self.forward(inputs)
        # logits = logits.reshape(self.num_estimators, -1, logits.size(-1))
        logits = rearrange(logits, "(n b) c -> b n c", n=self.num_estimators)
        probs_per_est = F.softmax(logits, dim=-1)
        probs = probs_per_est.mean(dim=1)
        confs, _ = probs.max(-1)

        if self.use_logits:
            ood_values, _ = -logits.mean(dim=1).max(dim=-1)
        elif self.use_entropy:
            ood_values = torch.special.entr(probs).sum(dim=-1).mean(dim=1)
        elif self.use_mi:
            mi_metric = MutualInformation(reduction="none")
            ood_values = mi_metric(probs_per_est)
        elif self.use_variation_ratio:
            vr_metric = VariationRatio(reduction="none", probabilistic=False)
            ood_values = vr_metric(probs_per_est.transpose(0, 1))
        else:
            ood_values = -confs

        if dataloader_idx == 0:
            self.test_cls_metrics.update(probs, targets)
            self.test_ood_metrics.update(ood_values, torch.zeros_like(targets))
            self.test_entropy_id(probs)
            self.test_id_ens_metrics.update(probs_per_est)
            self.log(
                "test_entropy_id",
                self.test_entropy_id,
                on_epoch=True,
                add_dataloader_idx=False,
            )
        else:
            self.test_ood_metrics.update(ood_values, torch.ones_like(targets))
            self.test_entropy_ood(probs)
            self.test_ood_ens_metrics.update(probs_per_est)
            self.log(
                "test_entropy_ood",
                self.test_entropy_ood,
                on_epoch=True,
                add_dataloader_idx=False,
            )

    def test_epoch_end(
        self, outputs: Union[EPOCH_OUTPUT, List[EPOCH_OUTPUT]]
    ) -> None:
        super().test_epoch_end(outputs)
        self.log_dict(
            self.test_id_ens_metrics.compute(),
        )
        self.log_dict(
            self.test_ood_ens_metrics.compute(),
        )
        self.test_id_ens_metrics.reset()
        self.test_ood_ens_metrics.reset()



# fmt: off
from typing import Literal, Optional

import torch
import torch.nn.functional as F
from torchmetrics import Metric
from torchmetrics.utilities.data import dim_zero_cat


# fmt:on
class BrierScore_old(Metric):
    """The Brier Score Metric.

    Args:
        reduction (str, optional): Determines how to reduce over the
            :math:`B`/batch dimension:

            - ``'mean'`` [default]: Averages score across samples
            - ``'sum'``: Sum score across samples
            - ``'none'`` or ``None``: Returns score per sample

        kwargs: Additional keyword arguments, see `Advanced metric settings
            <https://torchmetrics.readthedocs.io/en/stable/pages/overview.html#metric-kwargs>`_.

    Inputs:
        - :attr:`probs`: :math:`(B, C)` or :math:`(B, N, C)`
        - :attr:`target`: :math:`(B)`

        where :math:`B` is the batch size, :math:`C` is the number of classes
        and :math:`N` is the number of estimators.

    Warning:
        Make sure that the probabilities in :attr:`probs` are normalized to sum
        to one.

    Raises:
        ValueError:
            If :attr:`reduction` is not one of ``'mean'``, ``'sum'``,
            ``'none'`` or ``None``.
    """

    is_differentiable: bool = False
    higher_is_better: Optional[bool] = False
    full_state_update: bool = False

    def __init__(
        self, reduction: Literal["mean", "sum", "none", None] = "mean", **kwargs
    ) -> None:
        super().__init__(**kwargs)

        allowed_reduction = ("sum", "mean", "none", None)
        if reduction not in allowed_reduction:
            raise ValueError(
                "Expected argument `reduction` to be one of ",
                f"{allowed_reduction} but got {reduction}",
            )

        self.reduction = reduction
        self.num_estimators = 1

        if self.reduction in ["mean", "sum"]:
            self.add_state(
                "values", default=torch.tensor(0.0), dist_reduce_fx="sum"
            )
        else:
            self.add_state("values", default=[], dist_reduce_fx="cat")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, probs: torch.Tensor, target: torch.Tensor) -> None:
        """Update state with prediction probabilities and targets.

        Args:
            probs (torch.Tensor): Probabilities from the model.
            target (torch.Tensor): Ground truth labels.
        """
        batch_size = probs.size(0)
        if len(probs.shape) == 3:
            self.num_estimators = probs.size(1)
            target = target.unsqueeze(-1)
            target = target.repeat(1, self.num_estimators)

        brier_score = F.mse_loss(
            probs, F.one_hot(target), reduction="none"
        ).sum(dim=-1)

        if self.reduction is None or self.reduction == "none":
            self.values.append(brier_score)
        else:
            self.values += brier_score.sum()
            self.total += batch_size

    def compute(self) -> torch.Tensor:
        """Compute the final Brier score based on inputs passed to
        ``update``.
        """
        values = dim_zero_cat(self.values)
        if self.reduction == "sum":
            return values.sum(dim=-1) / self.num_estimators
        elif self.reduction == "mean":
            return values.sum(dim=-1) / self.total / self.num_estimators
        else:  # reduction is None
            return values


class BrierScore(Metric):
    full_state_update: bool = False

    def __init__(
        self, reduction: Literal["mean", "sum", "none", None] = "mean", **kwargs
    ) -> None:
        super().__init__(**kwargs)

        allowed_reduction = ("sum", "mean", "none", None)
        if reduction not in allowed_reduction:
            raise ValueError(
                "Expected argument `reduction` to be one of ",
                f"{allowed_reduction} but got {reduction}",
            )

        self.reduction = reduction
        self.num_estimators = 1

        if self.reduction in ["mean", "sum"]:
            self.add_state(
                "values", default=torch.tensor(0.0), dist_reduce_fx="sum"
            )
        else:
            self.add_state("values", default=[], dist_reduce_fx="cat")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, probs: torch.Tensor, target: torch.Tensor) -> None:
        """
        Update the current Brier score with a new tensor of probabilities.

        Args:
            probs (torch.Tensor): A probability tensor of shape
                (num_estimators, batch, num_classes) or
                (batch, num_classes)
        """
        if len(probs.shape) == 2:
            batch_size = probs.size(0)
        else:
            batch_size = probs.size(1)
            self.num_estimators = probs.size(2)
            target = target.repeat(1, self.num_estimators, 1)

        brier_score = F.mse_loss(probs, target, reduction="none").sum(dim=-1)

        if self.reduction is None or self.reduction == "none":
            self.values.append(brier_score)
        else:
            self.values += brier_score.sum()
            self.total += batch_size

    def compute(self) -> torch.Tensor:
        """
        Compute the final Brier score based on inputs passed to ``update``.

        Returns:
            torch.Tensor: The final value(s) for the Brier score
        """
        values = dim_zero_cat(self.values)
        if self.reduction == "sum":
            return values.sum(dim=-1) / self.num_estimators
        elif self.reduction == "mean":
            return values.sum(dim=-1) / self.total / self.num_estimators
        else:  # reduction is None
            return values


if __name__ == "__main__":
    input = torch.softmax(torch.randn(8, 4), dim=-1)
    target = torch.randint(4, (8,))

    bs_old = BrierScore_old()
    bs = BrierScore()

    assert bs(input, target) == bs_old(input, target)

    input = torch.softmax(torch.randn(8, 5, 4), dim=-1)

    assert bs(input, target) == bs_old(input.transpose(0, 1), target)

    print("All good!")



# flake8: noqa
from .brier_score import BrierScore
from .disagreement import Disagreement
from .entropy import Entropy
from .fpr95 import FPR95
from .jensen_shannon_divergence import JensenShannonDivergence
from .mutual_information import MutualInformation
from .nll import NegativeLogLikelihood
from .variation_ratio import VariationRatio



# fmt: off
from argparse import ArgumentParser
from typing import Any

import torch
import torch.nn as nn

from torch_uncertainty.models.resnet import (
    resnet18,
    resnet34,
    resnet50,
    resnet101,
    resnet152,
)
from torch_uncertainty.routines.classification import ClassificationSingle

# fmt: on
archs = [resnet18, resnet34, resnet50, resnet101, resnet152]
choices = [18, 34, 50, 101, 152]


class ResNet(ClassificationSingle):
    def __init__(
        self,
        num_classes: int,
        in_channels: int,
        arch: int,
        loss: nn.Module,
        optimization_procedure: Any,
        groups: int = 1,
        use_entropy: bool = False,
        use_logits: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(
            num_classes=num_classes,
            use_entropy=use_entropy,
            use_logits=use_logits,
        )

        self.save_hyperparameters(ignore=["loss", "optimization_procedure"])
        assert groups >= 1

        self.loss = loss
        self.optimization_procedure = optimization_procedure

        self.model = archs[choices.index(arch)](
            in_channels=in_channels,
            num_classes=num_classes,
            groups=groups,
        )

        # to log the graph
        self.example_input_array = torch.randn(1, in_channels, 32, 32)

    def configure_optimizers(self) -> dict:
        return self.optimization_procedure(self)

    @property
    def criterion(self) -> nn.Module:
        return self.loss()

    def forward(self, input: torch.Tensor) -> torch.Tensor:  # type: ignore
        return self.model.forward(input)

    @staticmethod
    def add_model_specific_args(
        parent_parser: ArgumentParser,
    ) -> ArgumentParser:
        parent_parser.add_argument(
            "--arch",
            type=int,
            default=18,
            choices=choices,
            help="Type of ResNet",
        )
        parent_parser.add_argument(
            "--entropy", dest="use_entropy", action="store_true"
        )
        parent_parser.add_argument("--groups", type=int, default=1)
        parent_parser.add_argument(
            "--logits", dest="use_logits", action="store_true"
        )
        return parent_parser



# fmt: off
from argparse import ArgumentParser
from pathlib import Path
from typing import Any, List, Optional, Union

import torchvision.transforms as T
from pytorch_lightning import LightningDataModule
from torch.utils.data import DataLoader, Dataset, random_split
from torchvision.datasets import SVHN

from ..datasets import TinyImageNet


# fmt: on
class TinyImageNetDataModule(LightningDataModule):
    def __init__(
        self,
        root: Union[str, Path],
        batch_size: int,
        val_split: int = 0,
        num_workers: int = 1,
        pin_memory: bool = True,
        persistent_workers: bool = True,
        **kwargs,
    ) -> None:
        super().__init__()
        # TODO: COMPUTE STATS
        if isinstance(root, str):
            root = Path(root)

        self.root: Path = root
        self.batch_size = batch_size
        self.val_split = val_split
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.persistent_workers = persistent_workers
        self.dataset = TinyImageNet
        self.ood_dataset = SVHN
        self.num_classes = 200

        self.transform_train = T.Compose(
            [
                T.RandomCrop(64, padding=4),
                T.RandomHorizontalFlip(),
                T.ToTensor(),
                T.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
            ]
        )

        self.transform_test = T.Compose(
            [
                T.Resize(64),
                T.ToTensor(),
                T.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
            ]
        )

    def _verify_splits(self, split: str) -> None:
        if split not in list(self.root.iterdir()):
            raise FileNotFoundError(
                f"a {split} TinyImagenet split was not found in {self.root},"
                f" make sure the folder contains a subfolder named {split}"
            )

    def prepare_data(self) -> None:
        # self.dataset(self.root, split="train")
        # self.dataset(self.root, split="val")
        self.ood_dataset(self.root, split="test", download=True)

    def setup(self, stage: Optional[str] = None) -> None:
        if stage == "fit" or stage is None:
            full = self.dataset(
                self.root,
                split="train",
                transform=self.transform_train,
            )
            self.train, self.val = random_split(
                full, [len(full) - self.val_split, self.val_split]
            )
            if self.val_split == 0:
                self.val = self.dataset(
                    self.root,
                    split="val",
                    transform=self.transform_test,
                )
        if stage == "test" or stage is None:
            self.test = self.dataset(
                self.root,
                split="val",
                transform=self.transform_test,
            )
            self.ood = self.ood_dataset(
                self.root,
                split="test",
                transform=self.transform_test,
            )

    def train_dataloader(self) -> DataLoader:
        return self._data_loader(self.train, shuffle=True)

    def val_dataloader(self) -> DataLoader:
        return self._data_loader(self.val)

    def test_dataloader(self) -> List[DataLoader]:
        r"""Gets test dataloaders for TinyImageNet.
        Returns:
            List[DataLoader]: TinyImageNet test set (in distribution data) and
            SVHN test split (out-of-distribution data).
        """
        return [self._data_loader(self.test), self._data_loader(self.ood)]

    def _data_loader(
        self, dataset: Dataset, shuffle: bool = False
    ) -> DataLoader:
        return DataLoader(
            dataset,
            batch_size=self.batch_size,
            shuffle=shuffle,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            persistent_workers=self.persistent_workers,
        )

    @classmethod
    def add_argparse_args(
        cls,
        parent_parser: ArgumentParser,
        **kwargs: Any,
    ) -> ArgumentParser:
        p = parent_parser.add_argument_group("datamodule")
        p.add_argument("--root", type=str, default="./data/")
        p.add_argument("--batch_size", type=int, default=256)
        p.add_argument("--val_split", type=int, default=0)
        p.add_argument("--num_workers", type=int, default=4)
        return parent_parser



"""_BatchedResNet in PyTorch.

Reference:
[1] Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun
    Deep Residual Learning for Image Recognition. arXiv:1512.03385
"""
# fmt: off
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from ...layers import BatchConv2d, BatchLinear

# fmt: on
__all__ = [
    "batched_resnet18",
    "batched_resnet34",
    "batched_resnet50",
    "batched_resnet101",
    "batched_resnet152",
]


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(
        self,
        in_planes: int,
        planes: int,
        stride: int = 1,
        num_estimators: int = 4,
    ) -> None:
        super().__init__()
        self.conv1 = BatchConv2d(
            in_planes,
            planes,
            kernel_size=3,
            num_estimators=num_estimators,
            stride=stride,
            padding=1,
            bias=False,
        )
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = BatchConv2d(
            planes,
            planes,
            kernel_size=3,
            num_estimators=num_estimators,
            stride=1,
            padding=1,
            bias=False,
        )
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(
                    in_planes,
                    self.expansion * planes,
                    kernel_size=1,
                    stride=stride,
                    bias=False,
                ),
                nn.BatchNorm2d(self.expansion * planes),
            )

    def forward(self, input: Tensor) -> Tensor:
        out = F.relu(self.bn1(self.conv1(input)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(input)
        out = F.relu(out)
        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(
        self,
        in_planes: int,
        planes: int,
        stride: int = 1,
        num_estimators: int = 4,
    ) -> None:
        super(Bottleneck, self).__init__()
        self.conv1 = BatchConv2d(
            in_planes,
            planes,
            kernel_size=1,
            num_estimators=num_estimators,
            bias=False,
        )
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = BatchConv2d(
            planes,
            planes,
            kernel_size=3,
            num_estimators=num_estimators,
            stride=stride,
            padding=1,
            bias=False,
        )
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = BatchConv2d(
            planes,
            self.expansion * planes,
            num_estimators=num_estimators,
            kernel_size=1,
            bias=False,
        )
        self.bn3 = nn.BatchNorm2d(self.expansion * planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                BatchConv2d(
                    in_planes,
                    self.expansion * planes,
                    kernel_size=1,
                    num_estimators=num_estimators,
                    stride=stride,
                    bias=False,
                ),
                nn.BatchNorm2d(self.expansion * planes),
            )

    def forward(self, input: Tensor) -> Tensor:
        out = F.relu(self.bn1(self.conv1(input)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(input)
        out = F.relu(out)
        return out


class _BatchedResNet(nn.Module):
    def __init__(
        self,
        block,
        num_blocks,
        in_channels: int,
        num_estimators,
        num_classes=10,
        width_multiplier: int = 1,
        # dataset: str = "cifar",
    ):
        super().__init__()
        self.in_planes = 64 * width_multiplier
        # self.dataset = dataset

        self.width_multiplier = width_multiplier
        # if dataset == "imagenet":
        #     self.conv1 = BatchConv2d(
        #         3,
        #         64 * self.width_multiplier,
        #         kernel_size=7,
        #         stride=2,
        #         padding=3,
        #         bias=False,
        #         num_estimators=num_estimators,
        #     )

        # elif dataset == "mnist":
        #     self.conv1 = BatchConv2d(
        #         1,
        #         64 * self.width_multiplier,
        #         kernel_size=3,
        #         stride=1,
        #         padding=1,
        #         bias=False,
        #         num_estimators=num_estimators,
        #     )
        # else:
        self.conv1 = BatchConv2d(
            in_channels,
            64 * self.width_multiplier,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=False,
            num_estimators=num_estimators,
        )
        self.bn1 = nn.BatchNorm2d(64 * self.width_multiplier)

        self.optional_pool = nn.Identity()

        self.layer1 = self._make_layer(
            block,
            64 * width_multiplier,
            num_blocks[0],
            stride=1,
            num_estimators=num_estimators,
        )
        self.layer2 = self._make_layer(
            block,
            128 * width_multiplier,
            num_blocks[1],
            stride=2,
            num_estimators=num_estimators,
        )
        self.layer3 = self._make_layer(
            block,
            256 * width_multiplier,
            num_blocks[2],
            stride=2,
            num_estimators=num_estimators,
        )
        self.layer4 = self._make_layer(
            block,
            512 * width_multiplier,
            num_blocks[3],
            stride=2,
            num_estimators=num_estimators,
        )
        self.pool = nn.AdaptiveAvgPool2d(output_size=1)
        self.flatten = nn.Flatten(1)

        self.linear = BatchLinear(
            512 * width_multiplier * block.expansion,
            num_classes,
            num_estimators=num_estimators,
        )

    def _make_layer(self, block, planes, num_blocks, stride, num_estimators):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride, num_estimators))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        # if self.dataset == "imagenet" or self.dataset == "tinyimagenet":
        #     out = F.max_pool2d(out, kernel_size=3, stride=2, padding=1)
        out = self.optional_pool(out)
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = self.pool(out)
        out = self.flatten(out)
        out = self.linear(out)
        return out


def batched_resnet18(
    in_channels: int, num_estimators: int, num_classes: int
) -> _BatchedResNet:
    """BatchEnsembles of ResNet-18 from `Deep Residual Learning for Image
    Recognition <https://arxiv.org/pdf/1512.03385.pdf>`_.

    Args:
        in_channels (int): Number of input channels.
        num_estimators (int): Number of estimators in the ensemble.
        groups (int): Number of groups within each estimator.
        num_classes (int): Number of classes to predict.

    Returns:
        _BatchedResNet: A BatchEnsembles-style ResNet-18.
    """
    return _BatchedResNet(
        BasicBlock,
        [2, 2, 2, 2],
        in_channels=in_channels,
        num_estimators=num_estimators,
        num_classes=num_classes,
    )


def batched_resnet34(
    in_channels: int, num_estimators: int, num_classes: int
) -> _BatchedResNet:
    """BatchEnsembles of ResNet-18 from `Deep Residual Learning for Image
    Recognition <https://arxiv.org/pdf/1512.03385.pdf>`_.

    Args:
        in_channels (int): Number of input channels.
        num_estimators (int): Number of estimators in the ensemble.
        groups (int): Number of groups within each estimator.
        num_classes (int): Number of classes to predict.

    Returns:
        _BatchedResNet: A BatchEnsembles-style ResNet-34.
    """
    return _BatchedResNet(
        BasicBlock,
        [3, 4, 6, 3],
        in_channels=in_channels,
        num_estimators=num_estimators,
        num_classes=num_classes,
    )


def batched_resnet50(
    in_channels: int,
    num_estimators: int,
    num_classes: int,
    width_multiplier: int = 1,
) -> _BatchedResNet:
    """BatchEnsembles of ResNet-18 from `Deep Residual Learning for Image
    Recognition <https://arxiv.org/pdf/1512.03385.pdf>`_.

    Args:
        in_channels (int): Number of input channels.
        num_estimators (int): Number of estimators in the ensemble.
        groups (int): Number of groups within each estimator.
        num_classes (int): Number of classes to predict.

    Returns:
        _BatchedResNet: A BatchEnsembles-style ResNet-50.
    """
    return _BatchedResNet(
        Bottleneck,
        [3, 4, 6, 3],
        in_channels=in_channels,
        num_estimators=num_estimators,
        num_classes=num_classes,
        width_multiplier=width_multiplier,
    )


def batched_resnet101(
    in_channels: int, num_estimators: int, num_classes: int
) -> _BatchedResNet:
    """BatchEnsembles of ResNet-18 from `Deep Residual Learning for Image
    Recognition <https://arxiv.org/pdf/1512.03385.pdf>`_.

    Args:
        in_channels (int): Number of input channels.
        num_estimators (int): Number of estimators in the ensemble.
        groups (int): Number of groups within each estimator.
        num_classes (int): Number of classes to predict.

    Returns:
        _BatchedResNet: A BatchEnsembles-style ResNet-101.
    """
    return _BatchedResNet(
        Bottleneck,
        [3, 4, 23, 3],
        in_channels=in_channels,
        num_estimators=num_estimators,
        num_classes=num_classes,
    )


def batched_resnet152(
    in_channels: int, num_estimators: int, num_classes: int
) -> _BatchedResNet:
    """BatchEnsembles of ResNet-18 from `Deep Residual Learning for Image
    Recognition <https://arxiv.org/pdf/1512.03385.pdf>`_.

    Args:
        in_channels (int): Number of input channels.
        num_estimators (int): Number of estimators in the ensemble.
        groups (int): Number of groups within each estimator.
        num_classes (int): Number of classes to predict.

    Returns:
        _BatchedResNet: A BatchEnsembles-style ResNet-152.
    """
    return _BatchedResNet(
        Bottleneck,
        [3, 8, 36, 3],
        in_channels=in_channels,
        num_estimators=num_estimators,
        num_classes=num_classes,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['torch_uncertainty',
 'torch_uncertainty.baselines',
 'torch_uncertainty.baselines.batched',
 'torch_uncertainty.baselines.masked',
 'torch_uncertainty.baselines.mimo',
 'torch_uncertainty.baselines.packed',
 'torch_uncertainty.baselines.standard',
 'torch_uncertainty.datamodules',
 'torch_uncertainty.datasets',
 'torch_uncertainty.layers',
 'torch_uncertainty.metrics',
 'torch_uncertainty.models',
 'torch_uncertainty.models.resnet',
 'torch_uncertainty.models.wideresnet',
 'torch_uncertainty.routines',
 'torch_uncertainty.transforms']

package_data = \
{'': ['*']}

install_requires = \
['einops>=0.6.0,<0.7.0',
 'pytorch-lightning>=1.9.0,<2.0.0',
 'scipy>=1.10.0,<2.0.0',
 'tensorboard>=2.11.2,<3.0.0',
 'timm>=0.6.12,<0.7.0',
 'torchinfo>=1.7.1,<2.0.0',
 'torchvision>=0.14.1,<0.15.0']

setup_kwargs = {
    'name': 'torch-uncertainty',
    'version': '0.1.0',
    'description': 'A PyTorch Library for benchmarking and leveraging efficient predictive uncertainty quantification techniques.',
    'long_description': '# Torch Uncertainty\n\n[![tests](https://github.com/ENSTA-U2IS/torch-uncertainty/actions/workflows/run-tests.yml/badge.svg?branch=main&event=push)](https://github.com/ENSTA-U2IS/torch-uncertainty/actions/workflows/run-tests.yml) [![Code Coverage](https://img.shields.io/codecov/c/github/ENSTA-U2IS/torch-uncertainty.svg)](https://codecov.io/gh/ENSTA-U2IS/torch-uncertainty) [![Code style: black](https://img.shields.io/badge/code%20style-black-black.svg)](https://github.com/psf/black)\n\n_Torch Uncertainty_ is a package designed to help you leverage uncertainty quantification techniques and make your neural networks more reliable. It is based on PyTorch Lightning to handle multi-GPU training and inference and automatic logging through tensorboard.\n\n---\n\nThis package provides a multi-level API, including:\n- ready-to-train baselines on research datasets, such as CIFAR and ImageNet\n- baselines available for training on your datasets\n- layers available for use in your networks\n\n## Installation\n\nThe package can be installed from PyPI or from source.\n\n### From PyPI (available soon)\n\nInstall the package via pip: `pip install torch-uncertainty`\n\n### From source\n\n#### Installing Poetry\n\nInstallation guidelines for poetry are available on <https://python-poetry.org/docs/>. They boil down to executing the following command:\n\n`curl -sSL https://install.python-poetry.org | python3 -`\n\n#### Installing the package\n\nClone the repository:\n\n`git clone https://github.com/ENSTA-U2IS/torch-uncertainty.git`\n\nCreate a new conda environment and activate it with:\n\n`conda create -n uncertainty && conda activate uncertainty`\n\nInstall the package using poetry:\n\n`poetry install torch-uncertainty` or, for development, `poetry install torch-uncertainty --with dev`\n\nDepending on your system, you may encounter errors. If so, kill the process and add `PYTHON_KEYRING_BACKEND=keyring.backends.null.Keyring` at the beginning of every `poetry install` commands.\n\n#### Contributing\n\nIn case that you would like to contribute, install from source and add the pre-commit hooks with `pre-commit install`\n\n## Getting Started and Documentation\n\nPlease find the documentation at [torch-uncertainty.github.io](https://torch-uncertainty.github.io).\n\nA quickstart is available at [torch-uncertainty.github.io/quickstart](https://torch-uncertainty.github.io/quickstart.html).\n\n## Implemented baselines\n\nTo date, the following baselines are implemented:\n\n- Deep Ensembles\n- Masksembles\n- Packed-Ensembles\n\n\n## Awesome Torch repositories\n\nYou may find a lot of information about modern uncertainty estimation techniques on the [Awesome Uncertainty in Deep Learning](https://github.com/ENSTA-U2IS/awesome-uncertainty-deeplearning).\n\n## References\n\n',
    'author': 'ENSTA U2IS',
    'author_email': 'olivier.laurent@ensta-paris.fr',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.12',
}


setup(**setup_kwargs)



# fmt: off
from argparse import ArgumentParser
from pathlib import Path
from typing import Any, List, Optional, Union

import torch
import torch.nn as nn
import torchvision.transforms as T
from pytorch_lightning import LightningDataModule
from timm.data.auto_augment import rand_augment_transform
from torch.utils.data import DataLoader, Dataset, random_split
from torchvision.datasets import CIFAR100, SVHN

from ..datasets import CIFAR100_C, AggregatedDataset
from ..transforms import Cutout


# fmt: on
class CIFAR100DataModule(LightningDataModule):
    def __init__(
        self,
        root: Union[str, Path],
        batch_size: int,
        val_split: int = 0,
        num_workers: int = 1,
        enable_cutout: bool = False,
        enable_randaugment: bool = False,
        auto_augment: str = None,
        use_cifar_c: str = None,
        corrupution_severity: int = 1,
        num_dataloaders: int = 1,
        pin_memory: bool = True,
        persistent_workers: bool = True,
        **kwargs,
    ) -> None:
        super().__init__()

        if isinstance(root, str):
            root = Path(root)

        self.root: Path = root
        self.batch_size = batch_size
        self.val_split = val_split
        self.num_workers = num_workers
        self.enable_cutout = enable_cutout
        self.enable_randaugment = enable_randaugment
        self.auto_augment = auto_augment
        self.num_dataloaders = num_dataloaders
        self.num_classes = 100
        self.num_channels = 3

        self.pin_memory = pin_memory
        self.persistent_workers = persistent_workers

        if use_cifar_c is None:
            self.dataset = CIFAR100
        else:
            self.dataset = CIFAR100_C

        self.ood_dataset = SVHN

        self.use_cifar_c = use_cifar_c
        self.corruption_severity = corrupution_severity

        assert (
            self.enable_cutout
            + self.enable_randaugment
            + int(self.auto_augment is not None)
            <= 1
        ), "Only one data augmentation can be chosen at a time."

        if enable_cutout:
            main_transform = Cutout(8)
        elif enable_randaugment:
            main_transform = T.RandAugment(num_ops=2, magnitude=20)
        elif auto_augment:
            main_transform = rand_augment_transform(self.auto_augment, {})
        else:
            main_transform = nn.Identity()

        self.transform_train = T.Compose(
            [
                T.RandomCrop(32, padding=4),
                T.RandomHorizontalFlip(),
                main_transform,
                T.ToTensor(),
                T.ConvertImageDtype(torch.float32),
                T.Normalize(
                    (0.5071, 0.4867, 0.4408),
                    (0.2675, 0.2565, 0.2761),
                ),
            ]
        )
        self.transform_test = T.Compose(
            [
                T.ToTensor(),
                T.Normalize(
                    (0.5071, 0.4867, 0.4408),
                    (0.2675, 0.2565, 0.2761),
                ),
            ]
        )

    def prepare_data(self) -> None:
        if self.use_cifar_c is None:
            self.dataset(self.root, train=True, download=True)
            self.dataset(self.root, train=False, download=True)

        self.ood_dataset(
            self.root,
            split="test",
            download=True,
            transform=self.transform_test,
        )

    def setup(self, stage: Optional[str] = None) -> None:
        if stage == "fit" or stage is None:
            assert (
                self.use_cifar_c is None
            ), "CIFAR-C can only be used in testing."
            full = self.dataset(
                self.root,
                train=True,
                download=False,
                transform=self.transform_train,
            )
            self.train, self.val = random_split(
                full, [len(full) - self.val_split, self.val_split]
            )
            if self.val_split == 0:
                self.val = self.dataset(
                    self.root,
                    train=False,
                    download=False,
                    transform=self.transform_test,
                )
        elif stage == "test":
            if self.use_cifar_c is None:
                self.test = self.dataset(
                    self.root,
                    train=False,
                    download=False,
                    transform=self.transform_test,
                )
            else:
                self.test = self.dataset(
                    self.root,
                    transform=self.transform_test,
                    subset=self.use_cifar_c,
                    severity=self.corruption_severity,
                )
            self.ood = self.ood_dataset(
                self.root,
                split="test",
                download=False,
                transform=self.transform_test,
            )

    def train_dataloader(self) -> DataLoader:
        r"""Gets the training dataloader for CIFAR10.
        Returns:
            DataLoader: CIFAR10 training dataloader.
        """
        if self.num_dataloaders > 1:
            return self._data_loader(
                AggregatedDataset(self.train, self.num_dataloaders),
                shuffle=True,
            )
        else:
            return self._data_loader(self.train, shuffle=True)

    def val_dataloader(self) -> DataLoader:
        r"""Gets the validation dataloader for CIFAR100.
        Returns:
            DataLoader: CIFAR100 validation dataloader.
        """
        return self._data_loader(self.val)

    def test_dataloader(self) -> List[DataLoader]:
        r"""Gets the test dataloaders for CIFAR100.
        Returns:
            List[DataLoader]: Dataloaders of the CIFAR100 test set (in
                distribution data) and SVHN test split (out-of-distribution
                data).
        """
        return [self._data_loader(self.test), self._data_loader(self.ood)]

    def _data_loader(
        self, dataset: Dataset, shuffle: bool = False
    ) -> DataLoader:
        return DataLoader(
            dataset,
            batch_size=self.batch_size,
            shuffle=shuffle,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            persistent_workers=self.persistent_workers,
        )

    @classmethod
    def add_argparse_args(
        cls,
        parent_parser: ArgumentParser,
        **kwargs: Any,
    ) -> ArgumentParser:
        p = parent_parser.add_argument_group("datamodule")
        p.add_argument("--root", type=str, default="./data/")
        p.add_argument("--batch_size", type=int, default=128)
        p.add_argument("--val_split", type=int, default=0)
        p.add_argument("--num_workers", type=int, default=4)
        p.add_argument("--cutout", dest="enable_cutout", action="store_true")
        p.add_argument(
            "--randaugment", dest="enable_randaugment", action="store_true"
        )
        p.add_argument("--auto_augment", type=str)
        p.add_argument("--cifar-c", dest="use_cifar_c", type=str, default=None)
        p.add_argument(
            "--severity", dest="corrupution_severity", type=int, default=1
        )
        return parent_parser



# fmt: off
import torch

import numpy as np

# fmt: on


class Cutout:
    """
    Reference : https://github.com/quark0/darts/blob/master/cnn/utils.py
    """

    def __init__(self, length: int):
        self.length = length

    def __call__(self, img: torch.Tensor):
        if self.length <= 0:
            return img
        h, w = img.size(1), img.size(2)
        mask = np.ones((h, w), np.float32)
        y = np.random.randint(h)
        x = np.random.randint(w)

        y1 = np.clip(y - self.length // 2, 0, h)
        y2 = np.clip(y + self.length // 2, 0, h)
        x1 = np.clip(x - self.length // 2, 0, w)
        x2 = np.clip(x + self.length // 2, 0, w)

        mask[y1:y2, x1:x2] = 0.0
        mask = torch.from_numpy(mask)
        mask = mask.expand_as(img)
        img *= mask
        return img



# fmt: off
from typing import List

import torch
from torch import Tensor
from torchmetrics import Metric
from torchmetrics.utilities import rank_zero_warn
from torchmetrics.utilities.data import dim_zero_cat

import numpy as np
from numpy.typing import ArrayLike


# fmt:on
def stable_cumsum(arr: ArrayLike, rtol: float = 1e-05, atol: float = 1e-08):
    """
    From https://github.com/hendrycks/anomaly-seg
    Uses high precision for cumsum and checks that the final value matches
    the sum.
    Args:
    arr : array-like
        To be cumulatively summed as flat
    rtol : float
        Relative tolerance, see ``np.allclose``
    atol : float
        Absolute tolerance, see ``np.allclose``
    """
    out = np.cumsum(arr, dtype=np.float64)
    expected = np.sum(arr, dtype=np.float64)
    if not np.allclose(out[-1], expected, rtol=rtol, atol=atol):
        raise RuntimeError(
            "cumsum was found to be unstable: "
            "its last element does not correspond to sum"
        )
    return out


class FPR95(Metric):
    """Class which computes the False Positive Rate at 95% Recall."""

    is_differentiable: bool = False
    higher_is_better: bool = False
    full_state_update: bool = False

    conf: List[Tensor]
    targets: List[Tensor]

    def __init__(self, pos_label: int, **kwargs) -> None:
        super().__init__(**kwargs)

        self.pos_label = pos_label
        self.add_state("conf", [], dist_reduce_fx="cat")
        self.add_state("targets", [], dist_reduce_fx="cat")

        rank_zero_warn(
            "Metric `FPR95` will save all targets and predictions"
            " in buffer. For large datasets this may lead to large memory"
            " footprint."
        )

    def update(self, conf: Tensor, target: Tensor) -> None:  # type: ignore
        self.conf.append(conf)
        self.targets.append(target)

    def compute(self) -> Tensor:
        r"""From https://github.com/hendrycks/anomaly-seg
        Compute the actual False Positive Rate at 95% Recall.
        Returns:
            Tensor: The value of the FPR95.
        """
        conf = dim_zero_cat(self.conf).cpu().numpy()
        targets = dim_zero_cat(self.targets).cpu().numpy()

        # out_labels is an array of 0s and 1s - 0 if IOD 1 if OOD
        out_labels = targets == self.pos_label

        in_scores = conf[np.logical_not(out_labels)]
        out_scores = conf[out_labels]

        # pos = OOD
        neg = np.array(in_scores[:]).reshape((-1, 1))
        pos = np.array(out_scores[:]).reshape((-1, 1))
        examples = np.squeeze(np.vstack((pos, neg)))
        labels = np.zeros(len(examples), dtype=np.int32)
        labels[: len(pos)] += 1

        # make labels a boolean vector, True if OOD
        labels = labels == self.pos_label

        # sort scores and corresponding truth values
        desc_score_indices = np.argsort(examples, kind="mergesort")[::-1]
        examples = examples[desc_score_indices]
        labels = labels[desc_score_indices]

        # examples typically has many tied values. Here we extract
        # the indices associated with the distinct values. We also
        # concatenate a value for the end of the curve.
        distinct_value_indices = np.where(np.diff(examples))[0]
        threshold_idxs = np.r_[distinct_value_indices, labels.shape[0] - 1]

        # accumulate the true positives with decreasing threshold
        tps = stable_cumsum(labels)[threshold_idxs]
        fps = 1 + threshold_idxs - tps  # add one because of zero-based indexing

        thresholds = examples[threshold_idxs]

        recall = tps / tps[-1]

        last_ind = tps.searchsorted(tps[-1])
        sl = slice(last_ind, None, -1)  # [last_ind::-1]
        recall, fps, tps, thresholds = (
            np.r_[recall[sl], 1],
            np.r_[fps[sl], 0],
            np.r_[tps[sl], 0],
            thresholds[sl],
        )

        cutoff = np.argmin(np.abs(recall - 0.95))

        return torch.as_tensor(
            fps[cutoff] / (np.sum(np.logical_not(labels))), dtype=torch.float32
        )



# fmt:off
from typing import Any, Literal, Optional

import torch
from torch import Tensor
from torchmetrics import Metric
from torchmetrics.utilities import rank_zero_warn
from torchmetrics.utilities.data import dim_zero_cat


# fmt:on
class MutualInformation(Metric):
    """The Mutual Information Metric to estimate the epistemic uncertainty of
    an ensemble of estimators.

    Args:
        reduction (str, optional): Determines how to reduce over the
            :math:`B`/batch dimension:

            - ``'mean'`` [default]: Averages score across samples
            - ``'sum'``: Sum score across samples
            - ``'none'`` or ``None``: Returns score per sample

        kwargs: Additional keyword arguments, see `Advanced metric settings
            <https://torchmetrics.readthedocs.io/en/stable/pages/overview.html#metric-kwargs>`_.

    Inputs:
        - :attr:`probs`: :math:`(B, N, C)`

        where :math:`B` is the batch size, :math:`C` is the number of classes
        and :math:`N` is the number of estimators.

    Note:
        A higher mutual information means a higher uncertainty.

    Warning:
        Make sure that the probabilities in :attr:`probs` are normalized to sum
        to one.

    Raises:
        ValueError:
            If :attr:`reduction` is not one of ``'mean'``, ``'sum'``,
            ``'none'`` or ``None``.
    """

    is_differentiable: bool = False
    higher_is_better: Optional[bool] = None
    full_state_update: bool = False

    def __init__(
        self,
        reduction: Literal["mean", "sum", "none", None] = "mean",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        allowed_reduction = ("sum", "mean", "none", None)
        if reduction not in allowed_reduction:
            raise ValueError(
                "Expected argument `reduction` to be one of ",
                f"{allowed_reduction} but got {reduction}",
            )

        self.reduction = reduction

        if self.reduction in ["mean", "sum"]:
            self.add_state(
                "values", default=torch.tensor(0.0), dist_reduce_fx="sum"
            )
        else:
            self.add_state("values", default=[], dist_reduce_fx="cat")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, probs: torch.Tensor) -> None:
        """Update the current mutual information with a new tensor of
        probabilities.

        Args:
            probs (torch.Tensor): Probabilities from the ensemble.
        """
        batch_size = probs.size(0)

        ens_probs = probs.mean(dim=1)
        entropy_mean = torch.special.entr(ens_probs).sum(dim=-1)
        mean_entropy = torch.special.entr(probs).sum(dim=-1).mean(dim=1)

        mutual_information = entropy_mean - mean_entropy

        if self.reduction is None or self.reduction == "none":
            self.values.append(mutual_information)
        else:
            self.values += mutual_information.sum()
            self.total += batch_size

    def compute(self) -> torch.Tensor:
        """Computes Mutual Information based on inputs passed in to ``update``
        previously.
        """
        values = dim_zero_cat(self.values)
        if self.reduction == "sum":
            return values.sum(dim=-1)
        elif self.reduction == "mean":
            return values.sum(dim=-1) / self.total
        else:  # reduction is None or "none"
            return values


class MutualInformation_old(Metric):
    r"""
    The Mutual Information to estimate the epistemic uncertainty.
    A higher mutual information means a higher uncertainty.
    """
    full_state_update: bool = False

    def __init__(
        self, reduction: Literal["mean", "sum", "none", None] = "mean", **kwargs
    ) -> None:
        super().__init__(**kwargs)

        allowed_reduction = ("sum", "mean", "none", None)
        if reduction not in allowed_reduction:
            raise ValueError(
                "Expected argument `reduction` to be one of ",
                f"{allowed_reduction} but got {reduction}",
            )

        self.reduction = reduction

        self.add_state("probs_per_est", [], dist_reduce_fx="cat")

        rank_zero_warn(
            "Metric `MutualInformation` will save and predictions "
            "in buffer. For large datasets this may lead to a large memory "
            "footprint."
        )

    def update(self, probs_per_est: Tensor) -> None:  # type: ignore
        # store data as (example, estimator, class)
        if len(probs_per_est.shape) <= 2:
            raise ValueError("Please give the probabilities per estimator.")

        self.probs_per_est.append(probs_per_est.transpose(0, 1))

    def compute(self) -> Tensor:
        r"""Computes the mutual information on the data.
        Returns:
            Tensor: The total mutual information.
        """

        # convert data to (estimator, example, class)
        probs_per_est = dim_zero_cat(self.probs_per_est).transpose(0, 1)
        probs = probs_per_est.mean(dim=0)

        # Entropy of the mean over the estimators
        entropy_mean = torch.special.entr(probs).sum(dim=-1)

        # Mean over the estimators of the entropy over the classes
        mean_entropy = torch.special.entr(probs_per_est).sum(dim=-1).mean(dim=0)

        mutual_information: torch.Tensor = entropy_mean - mean_entropy

        if self.reduction == "mean":
            mutual_information = mutual_information.mean()
        elif self.reduction == "sum":
            mutual_information = mutual_information.sum()

        return mutual_information


if __name__ == "__main__":
    input = torch.softmax(torch.randn(8, 5, 4), dim=-1)  # (B, N, C)

    mi_old = MutualInformation_old()
    mi = MutualInformation()

    assert mi_old(input.transpose(0, 1)) == mi(input)

    print("All good!")



# fmt: off
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange

from ...layers import PackedConv2d, PackedLinear


# fmt: on
class WideBasicBlock(nn.Module):
    def __init__(
        self,
        in_planes,
        planes,
        dropout_rate,
        stride=1,
        num_estimators=4,
        gamma=1,
    ):
        super().__init__()
        self.bn1 = nn.BatchNorm2d(in_planes)
        self.conv1 = PackedConv2d(
            in_planes,
            planes,
            kernel_size=3,
            num_estimators=num_estimators,
            padding=1,
            groups=gamma,
            bias=False,
        )
        self.dropout = nn.Dropout(p=dropout_rate)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv2 = PackedConv2d(
            planes,
            planes,
            kernel_size=3,
            num_estimators=num_estimators,
            stride=stride,
            padding=1,
            groups=gamma,
            bias=False,
        )
        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != planes:
            self.shortcut = nn.Sequential(
                PackedConv2d(
                    in_planes,
                    planes,
                    kernel_size=1,
                    num_estimators=num_estimators,
                    stride=stride,
                    groups=gamma,
                    bias=True,
                ),
            )

    def forward(self, x):
        out = self.dropout(self.conv1(F.relu(self.bn1(x))))
        out = self.conv2(F.relu(self.bn2(out)))
        out += self.shortcut(x)
        return out


class _PackedWide(nn.Module):
    def __init__(
        self,
        depth: int,
        widen_factor: int,
        in_channels: int,
        num_classes: int,
        num_estimators: int,
        dropout_rate: float,
        alpha: int = 2,
        gamma: int = 1,
    ):
        super().__init__()
        self.num_estimators = num_estimators
        self.in_planes = 16 * alpha

        assert (depth - 4) % 6 == 0, "Wide-resnet depth should be 6n+4"
        n = (depth - 4) / 6
        k = widen_factor

        nStages = [16, 16 * k, 32 * k, 64 * k]

        self.conv1 = PackedConv2d(
            in_channels * self.num_estimators,
            nStages[0] * alpha,
            kernel_size=3,
            num_estimators=self.num_estimators,
            stride=1,
            padding=1,
            groups=gamma,
            bias=True,
        )

        self.optional_pool = nn.Identity()

        self.layer1 = self._wide_layer(
            WideBasicBlock,
            nStages[1] * alpha,
            n,
            dropout_rate,
            stride=1,
            num_estimators=self.num_estimators,
            gamma=gamma,
        )
        self.layer2 = self._wide_layer(
            WideBasicBlock,
            nStages[2] * alpha,
            n,
            dropout_rate,
            stride=2,
            num_estimators=self.num_estimators,
            gamma=gamma,
        )
        self.layer3 = self._wide_layer(
            WideBasicBlock,
            nStages[3] * alpha,
            n,
            dropout_rate,
            stride=2,
            num_estimators=self.num_estimators,
            gamma=gamma,
        )
        self.bn1 = nn.BatchNorm2d(nStages[3] * alpha, momentum=0.9)

        self.pool = nn.AdaptiveAvgPool2d(output_size=1)
        self.flatten = nn.Flatten(1)

        self.linear = PackedLinear(
            nStages[3] * alpha,
            num_classes * num_estimators,
            num_estimators,
        )

    def _wide_layer(
        self,
        block: nn.Module,
        planes: int,
        num_blocks: int,
        dropout_rate: float,
        stride: int,
        num_estimators: int,
        gamma,
    ):
        strides = [stride] + [1] * (int(num_blocks) - 1)
        layers = []

        for stride in strides:
            layers.append(
                block(
                    self.in_planes,
                    planes,
                    dropout_rate,
                    stride,
                    num_estimators,
                    gamma,
                )
            )
            self.in_planes = planes

        return nn.Sequential(*layers)

    def forward(self, x):
        out = self.conv1(x)
        out = self.optional_pool(out)
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = F.relu(self.bn1(out))
        out = rearrange(
            out, "e (m c) h w -> (m e) c h w", m=self.num_estimators
        )
        out = self.pool(out)
        out = self.flatten(out)
        out = self.linear(out)

        return out


def packed_wideresnet28x10(
    num_estimators: int, alpha: int, gamma: int, num_classes: int
) -> nn.Module:
    return _PackedWide(
        depth=28,
        widen_factor=10,
        dropout_rate=0.3,
        num_classes=num_classes,
        num_estimators=num_estimators,
        alpha=alpha,
        gamma=gamma,
    )



# flake8: noqa
from .aggregated_dataset import AggregatedDataset
from .cifar_c import CIFAR10_C, CIFAR100_C
from .imagenet_o import ImageNetO
from .imagenet_r import ImageNetR
from .tiny_imagenet import TinyImageNet



# fmt: off
from typing import Any, List, Literal, Optional

import torch
import torch.nn.functional as F
from torch import Tensor
from torchmetrics import Metric
from torchmetrics.utilities import rank_zero_warn
from torchmetrics.utilities.data import dim_zero_cat


# fmt: on
class Disagreement(Metric):
    """The Disagreement Metric to estimate the confidence of an ensemble of
    estimators.

    Args:
        reduction (str, optional): Determines how to reduce over the
            :math:`B`/batch dimension:

            - ``'mean'`` [default]: Averages score across samples
            - ``'sum'``: Sum score across samples
            - ``'none'`` or ``None``: Returns score per sample

        kwargs: Additional keyword arguments, see `Advanced metric settings
            <https://torchmetrics.readthedocs.io/en/stable/pages/overview.html#metric-kwargs>`_.

    Inputs:
        - :attr:`probs`: :math:`(B, N, C)`

        where :math:`B` is the batch size, :math:`C` is the number of classes
        and :math:`N` is the number of estimators.

    Note:
        A higher disagreement means a lower confidence.

    Warning:
        Make sure that the probabilities in :attr:`probs` are normalized to sum
        to one.

    Raises:
        ValueError:
            If :attr:`reduction` is not one of ``'mean'``, ``'sum'``,
            ``'none'`` or ``None``.
    """

    is_differentiable: bool = False
    higher_is_better: Optional[bool] = None
    full_state_update: bool = False

    def __init__(
        self,
        reduction: Literal["mean", "sum", "none", None] = "mean",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        allowed_reduction = ("sum", "mean", "none", None)
        if reduction not in allowed_reduction:
            raise ValueError(
                "Expected argument `reduction` to be one of ",
                f"{allowed_reduction} but got {reduction}",
            )

        self.reduction = reduction

        if self.reduction in ["mean", "sum"]:
            self.add_state(
                "values",
                default=torch.tensor(0.0),
                dist_reduce_fx="sum",
            )
        else:
            self.add_state("values", default=[], dist_reduce_fx="cat")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

    def _compute_disagreement(self, preds: Tensor) -> Tensor:
        num_estimators = preds.size(-1)
        counts = torch.sum(F.one_hot(preds), dim=1)
        max_counts = num_estimators * (num_estimators - 1) / 2
        return 1 - (counts * (counts - 1) / 2).sum(dim=1) / max_counts

    def update(self, probs: Tensor) -> None:
        """Update state with prediction probabilities and targets.

        Args:
            probs (torch.Tensor): Probabilities from the model.
        """
        preds = probs.argmax(dim=-1)
        if self.reduction is None or self.reduction == "none":
            self.values.append(self._compute_disagreement(preds))
        else:
            self.values += self._compute_disagreement(preds).sum(dim=-1)
            self.total += probs.size(0)

    def compute(self) -> Tensor:
        """Compute Disagreement based on inputs passed in to ``update``."""
        values = dim_zero_cat(self.values)
        if self.reduction == "sum":
            return values.sum(dim=-1)
        elif self.reduction == "mean":
            return values.sum(dim=-1) / self.total
        else:  # reduction is None or "none"
            return values


# TODO: Check dimension
class Disagreement_old(Metric):
    full_state_update: bool = False
    probs: List[Tensor]

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.add_state("probs", [], dist_reduce_fx="cat")

        rank_zero_warn(
            "Metric `Disagreement` will save all targets and predictions"
            " in buffer. For large datasets this may lead to large memory"
            " footprint."
        )

    def update(self, probs: Tensor) -> None:  # type: ignore
        # store data as (example, estimator, class)
        self.probs.append(probs.transpose(0, 1))

    def _compute_disagreement(self, classes: Tensor) -> Tensor:
        r"""Computes the disagreement between the predicted classes among
        all pairs of estimators.
        Args:
            classes (Tensor): Classes predicted by the `n_estimators`
                estimators.
        Returns:
            Tensor: Mean disagreement between estimators.
        """
        # TODO: Using onehot might be memory intensive
        n_estimators = classes.shape[-1]
        counts = torch.sum(F.one_hot(classes), dim=1)
        max_counts = n_estimators * (n_estimators - 1) / 2
        return torch.mean(
            1 - (counts * (counts - 1) / 2).sum(dim=1) / max_counts
        )

    def compute(self) -> Tensor:
        probs = dim_zero_cat(self.probs)
        classes = probs.argmax(dim=-1)
        return self._compute_disagreement(classes)


if __name__ == "__main__":
    input = torch.softmax(torch.randn(8, 5, 4), dim=-1)  # (B, N, C)

    disagreement_old = Disagreement_old()
    disagreement = Disagreement()

    assert disagreement_old(input.transpose(0, 1)) == disagreement(input)

    print("All good!")
